#!/usr/bin/env python
"""
tap protocol client.

Copyright (c) 2010  Dustin Sallings <dustin@spy.net>
"""
import exceptions

import sys
import socket
import string
import random
import struct
import asyncore
import uuid
import mc_bin_server

from memcacheConstants import REQ_MAGIC_BYTE
from memcacheConstants import REQ_PKT_FMT

import memcacheConstants

class TapConnection(object):
    def __init__(self, server, port, callback, clientId=None, opts={}):
        self.wbuf = self._createTapCall(clientId, opts)
        print self.wbuf
        self.callback = callback
        self.identifier = (server.ip, port)
        print 'tap connection : {0} {1}'.format(server.ip, port)
        self.s = socket.socket()
        self.s.connect_ex((server.ip, port))
        sent = self.s.send(self.wbuf)
        self.wbuf = self.wbuf[sent:]
        self.rbuf = ""


    def __hasEnoughBytes(self):
        rv = False
        if len(self.rbuf) >= memcacheConstants.MIN_RECV_PACKET:
            magic, cmd, keylen, extralen, datatype, vb, remaining, opaque, cas =\
            struct.unpack(REQ_PKT_FMT, self.rbuf[:memcacheConstants.MIN_RECV_PACKET])
            rv = len(self.rbuf) - memcacheConstants.MIN_RECV_PACKET >= remaining
        return rv

    def receive(self):
        BUFFER_SIZE = 4096
        self.s.settimeout(10)
        try:
            self.rbuf += self.s.recv(BUFFER_SIZE)
        except:
            return
        while self.__hasEnoughBytes():
            magic, cmd, keylen, extralen, datatype, vb, remaining, opaque, cas =\
            struct.unpack(REQ_PKT_FMT, self.rbuf[:memcacheConstants.MIN_RECV_PACKET])
            assert magic == REQ_MAGIC_BYTE
            assert keylen <= remaining, "Keylen is too big: %d > %d"\
            % (keylen, remaining)
            assert extralen == memcacheConstants.EXTRA_HDR_SIZES.get(cmd, 0),\
            "Extralen is too large for cmd 0x%x: %d" % (cmd, extralen)
            # Grab the data section of this request
            data = self.rbuf[memcacheConstants.MIN_RECV_PACKET:memcacheConstants.MIN_RECV_PACKET + remaining]
            assert len(data) == remaining
            # Remove this request from the read buffer
            self.rbuf = self.rbuf[memcacheConstants.MIN_RECV_PACKET + remaining:]
            # Process the command
            cmdVal = self.processCommand(cmd, keylen, vb, extralen, cas, data)
            # Queue the response to the client if applicable.
            if cmdVal:
                try:
                    status, cas, response = cmdVal
                except ValueError:
                    print "Got", cmdVal
                    raise
                dtype = 0
                extralen = memcacheConstants.EXTRA_HDR_SIZES.get(cmd, 0)
                self.wbuf += struct.pack(memcacheConstants.RES_PKT_FMT,
                                         memcacheConstants.RES_MAGIC_BYTE, cmd, keylen,
                                         extralen, dtype, status,
                                         len(response), opaque, cas) + response


    def _createTapCall(self, key=None, opts={}):
        # Client identifier
        if not key:
            key = "".join(random.sample(string.letters, 16))
        dtype = 0
        opaque = 0
        cas = 0

        extraHeader, val = self._encodeOpts(opts)
        print "opts: ", extraHeader, val

        msg = struct.pack(REQ_PKT_FMT, REQ_MAGIC_BYTE,
                          memcacheConstants.CMD_TAP_CONNECT,
                          len(key), len(extraHeader), dtype, 0,
                          len(key) + len(extraHeader) + len(val),
                          opaque, cas)
        return msg + extraHeader + key + val

    def _encodeOpts(self, opts):
        header = 0
        val = []
        for op in sorted(opts.keys()):
            header |= op
            if op in memcacheConstants.TAP_FLAG_TYPES:
                val.append(struct.pack(memcacheConstants.TAP_FLAG_TYPES[op],
                                       opts[op]))
            elif op == memcacheConstants.TAP_FLAG_LIST_VBUCKETS:
                val.append(self._encodeVBucketList(opts[op]))
            else:
                val.append(opts[op])
        return struct.pack(">I", header), ''.join(val)

    def _encodeVBucketList(self, vbl):
        l = list(vbl) # in case it's a generator
        vals = [struct.pack("!H", len(l))]
        for v in vbl:
            vals.append(struct.pack("!H", v))
        return ''.join(vals)

    def processCommand(self, cmd, klen, vb, extralen, cas, data):
        extra = data[0:extralen]
        key = data[extralen:(extralen + klen)]
        val = data[(extralen + klen):]
        self.callback(self.identifier, cmd, extra, key, vb, val, cas)

    def handle_connect(self):
        print "connected..."


    def handle_close(self):
        print "handle_close"
        self.close()


class TapClient(object):
    def __init__(self, servers, callback, opts={}):
        for t in servers:
            tc = TapConnection(t.host, t.port, callback, t.id, opts)


def buildGoodSet(goodChars=string.printable, badChar='?'):
    """Build a translation table that turns all characters not in goodChars
    to badChar"""
    allChars = string.maketrans("", "")
    badchars = string.translate(allChars, allChars, goodChars)
    rv = string.maketrans(badchars, badChar * len(badchars))
    return rv


class TapDescriptor(object):
    port = 11210
    id = None

    def __init__(self, server):
        self.host = server.ip
        self.port = 11210
        self.id = str(uuid.uuid4())

# Build a translation table that includes only characters
transt = buildGoodSet()

def abbrev(v, maxlen=30):
    if len(v) > maxlen:
        return v[:maxlen] + "..."
    else:
        return v


def keyprint(v):
    return string.translate(abbrev(v), transt)

#def mainLoop(serverList, cb, opts={}):
#    """Run the given callback for each tap message from any of the
#    upstream servers.
#
#    loops until all connections drop
#    """
#
#    connections = (TapDescriptor(a) for a in serverList)
#    TapClient(connections, cb, opts=opts)
#    asyncore.loop()

#if __name__ == '__main__':
#    def cb(identifier, cmd, extra, key, vb, val, cas):
#        print "%s: ``%s'' (vb:%d) -> ``%s'' (%d bytes from %s)" % (
#            memcacheConstants.COMMAND_NAMES[cmd],
#            key, vb, keyprint(val), len(val), identifier)
#
#    # This is an example opts parameter to do future-only tap:
#    opts = {memcacheConstants.TAP_FLAG_BACKFILL: 0xffffffff}
#    # If you omit it, or supply a past time_t value for backfill, it
#    # will get all data.
#    opts = {}
#
#    mainLoop(sys.argv[1:], cb, opts)




from old_tasks.task import NodeInitializeTask, BucketCreateTask

class TaskHelper():
    @staticmethod
    def add_node_init_task(tm, server):
        node_init_task = NodeInitializeTask(server)
        tm.schedule(node_init_task)
        return node_init_task

    @staticmethod
    def add_bucket_create_task(tm, server, bucket='default', replicas=1, port=11210, size=0,
                           password=None):
        bucket_create_task = BucketCreateTask(server, bucket, replicas, port, size, password)
        tm.schedule(bucket_create_task)
        return bucket_create_task



import xml.dom.minidom

# a junit compatible xml example
#<?xml version="1.0" encoding="UTF-8"?>
#<testsuite name="nosetests" tests="1" errors="1" failures="0" skip="0">
#    <testcase classname="path_to_test_suite.TestSomething"
#              name="path_to_test_suite.TestSomething.test_it" time="0">
#        <error type="exceptions.TypeError">
#        Traceback (most recent call last):
#        ...
#        TypeError: oops, wrong type
#        </error>
#    </testcase>
#</testsuite>

#
# XUnitTestCase has name , time and error
# error is a XUnitTestCase
class XUnitTestCase(object):
    def __init__(self):
        self.name = ""
        self.time = 0
        self.error = None
        self.params = ''
#
# XUnitTestCaseError has type and message
#
class XUnitTestCaseError(object):
    def __init__(self):
        self.type = ""
        self.message = ""


#
# XUnitTestSuite has name , time , list of XUnitTestCase objects
# errors : number of errors
# failures : number of failures
# skips : number of skipped tests

#

class XUnitTestResult(object):

    def __init__(self):
        self.suites = []

    def add_test(self, name, time=0, errorType=None, errorMessage=None, status='pass', params=''):
        #Get the classname
        class_name = name[:name.rfind(".")]
        # If params are passed to the test
        if "," in name:
            class_name = name

        matched = False
        for suite in self.suites:
            if suite.name == class_name:
                suite.add_test(name, time, errorType, errorMessage, status, params=params)
                matched = True
                break
        if not matched:
            suite = XUnitTestSuite()
            suite.name = class_name
            suite.add_test(name, time, errorType, errorMessage, status, params=params)
            self.suites.append(suite)

    def to_xml(self, suite):
        doc = xml.dom.minidom.Document()
        testsuite = doc.createElement('testsuite')
        #<testsuite name="nosetests" tests="1" errors="1" failures="0" skip="0">
        testsuite.setAttribute('name', suite.name)
        testsuite.setAttribute('errors', str(suite.errors))
        testsuite.setAttribute('failures', str(suite.failures))
        testsuite.setAttribute('errors', str(suite.errors))
        testsuite.setAttribute('tests', str(len(suite.tests)))
        testsuite.setAttribute('time', str(suite.time))
        testsuite.setAttribute('skip', str(suite.skips))
        for testobject in suite.tests:
            testcase = doc.createElement('testcase')
            full_name = testobject.name+testobject.params
            testcase.setAttribute('name', full_name)
            testcase.setAttribute('time', str(testobject.time))
            if testobject.error:
                error = doc.createElement('error')
                error.setAttribute('type', testobject.error.type)
                if testobject.error.message:
                    message = doc.createTextNode(testobject.error.message)
                    error.appendChild(message)
                testcase.appendChild(error)
            testsuite.appendChild(testcase)
        doc.appendChild(testsuite)
        return doc.toprettyxml()

    def write(self, prefix, params=''):
        for suite in self.suites:
            name = suite.name
            # If test_params are passed
            if "," in name:
                name = name[:name.find(",")]

            report_xml_file = open("{0}-{1}.xml".format(prefix, name), 'w')
            report_xml_file.write(self.to_xml(suite))
            report_xml_file.close()

    def print_summary(self):
        for suite in self.suites:
            oks = []
            errors = []
            for test in suite.tests:
                if test.error:
                    errors.append(test.name)
                else:
                    oks.append(test.name)
            msg = "summary so far suite {0} , pass {1} , fail {2}"
            print msg.format(suite.name, len(oks), len(errors))
            if errors:
                print "failures so far..."
                for error in errors:
                    print error


class XUnitTestSuite(object):
    def __init__(self):
        self.name = ""
        self.time = 0
        self.tests = []
        self.errors = 0
        self.failures = 0
        self.skips = 0

    # create a new XUnitTestCase and update the errors/failures/skips count
    def add_test(self, name, time=0, errorType=None, errorMessage=None, status='pass', params=''):
        #create a test_case and add it to this suite
        # todo: handle 'skip' or 'setup_failure' or other
        # status codes that testrunner might pass to this function
        test = XUnitTestCase()
        test.name = name
        test.time = time
        test.params = params
        if status == 'fail':
            error = XUnitTestCaseError()
            error.type = errorType
            error.message = errorMessage
            test.error = error
        self.tests.append(test)
        if status == 'fail':
            self.failures += 1
            self.errors += 1
        elif status == 'skip':
            self.skips += 1
        self.time += time


    # generate the junit xml representation from the XUnitTestSuite object
    # todo : create an element for errorMessage and append it to to error node



#!/usr/bin/env python

import logging
import os
import json
from source import Source
from server import Server

from lib.memcached.helper.data_helper import MemcachedClientHelper

class MemcachedSource(Source):

    META_FILE = "%s/../metadata/stats.json" % os.path.dirname(__file__)
    MC_STATS = ["", "allocator", "checkpoint", "config", "dispatcher",
                "hash", "kvstore", "kvtimings", "memory", "prev-vbucket",
                "tap", "tapagg", "timings", "vbucket", "vbucket-details"]

    def __init__(self, server, bucket, meta_file=META_FILE, mc=None):
        self.server = server        # @class: Server
        self.bucket = bucket
        self.mc = mc
        self.data = {}
        self.meta = self.get_meta(meta_file)
        self.slow = {"mc-host": server.ip, "mc-bucket": bucket}
        self.fast = {"mc-host": server.ip, "mc-bucket": bucket}

    def get_meta(self, meta_file):
        with open(meta_file) as f:
            return json.loads(f.read())

    def filter(self, stats):
        if not stats or not self.meta:
            logging.error("unable to apply filter: invalid data or meta")
            return False

        for key in stats:
            mkey = "mc-%s" % key
            if mkey in self.meta and "visit" in self.meta[mkey]:
                if self.meta[mkey]["visit"] == "slow":
                    self.slow[mkey] = stats[key]
                elif self.meta[mkey]["visit"] == "fast":
                    self.fast[mkey] = stats[key]

    def connect(self):
        if not self.server or \
            not isinstance(self.server, Server):
            return False

        self.mc = MemcachedClientHelper.direct_client(
            self.server, self.bucket)

        return True

    def feed(self):
        if not self.mc and not self.connect():
            logging.error("unable to connect to server : %s" % self.server)
            return None

        self.data = {}
        for key in MemcachedSource.MC_STATS:
            try:
                stats = self.mc.stats(key)
                if key != "":
                    self.data[key] = stats
                else:
                    self.data["all"] = stats
                self.filter(stats)
            except Exception, e:
                logging.error("exception for key %s : %s" % (key, e))

        return self.data

    def disconnect(self):
        self.mc.close()

    def gen_stats(self):
        """
        Generate individual stats
        """
        for data in self.data.itervalues():
            for key_val in data.iteritems():
                yield key_val


#!/usr/bin/env python

import logging
from functools import wraps

class SocketHelper:

    @staticmethod
    def send_bytes(skt, buf, timeout=0):
        """
        Send bytes to the socket
        @raise ValueError, IOError, socket.timeout, Exception
        """
        if not buf or not skt:
            raise ValueError("<send_bytes> invalid socket descriptor or buf")

        if timeout:
            skt.settimeout(timeout)
        else:
            skt.settimeout(None)

        length = len(buf)
        sent_total = 0

        while sent_total < length:
            sent = skt.send(buf)
            if not sent:
                raise IOError("<send_bytes> connection broken")
            sent_total += sent
            buf = buf[sent:]

        return sent_total

    @staticmethod
    def recv_bytes(skt, length, timeout=0):
        """
        Receive bytes from the socket
        @raise ValueError, IOError, socket.timeout, Exception
        """
        if not skt or length < 0:
            raise ValueError("<recv_bytes> invalid socket descriptor or length")

        if timeout:
            skt.settimeout(timeout)
        else:
            skt.settimeout(None)

        buf = ''
        while len(buf) < length:
            data = skt.recv(length - len(buf))
            if not data:
                raise IOError("<recv_bytes> connection broken")
            buf += data

        return buf

    @staticmethod
    def send_bytes_ex(skt, buf, timeout=0):
        """
        Send bytes to the socket. Swallow exceptions.
        """
        try:
            return SocketHelper.send_bytes(skt, buf, timeout)
        except Exception, e:
            logging.error(e)

        return -1

    @staticmethod
    def recv_bytes_ex(skt, length, timeout=0):
        """
        Receive bytes to the socket. Swallow exceptions.
        """
        try:
            return SocketHelper.recv_bytes(skt, length, timeout)
        except Exception, e:
            logging.error(e)

        return None

def synchronized(lock_name):
    """synchronized access to class method"""
    def _outer(func):
        @wraps(func)
        def _inner(self, *args, **kwargs):
            try:
                lock = self.__getattribute__(lock_name)
            except AttributeError:
                logging.error("<synchronized> cannot find lock: %s", lock_name)
                return _inner
            with lock:
                return func(self, *args, **kwargs)
        return _inner
    return _outer

def is_num(val):
    """check if a string is numeric"""
    try:
        float(val)
    except ValueError:
        return False

    return True


# Copyright 2009-2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from uuid import UUID
except ImportError:
    # Python2.4 doesn't have a uuid module.
    pass

"""Tools for representing BSON binary data.
"""

BINARY_SUBTYPE = 0
"""BSON binary subtype for binary data.

This is the default subtype for binary data.

.. versionadded:: 1.5
"""

FUNCTION_SUBTYPE = 1
"""BSON binary subtype for functions.

.. versionadded:: 1.5
"""

OLD_BINARY_SUBTYPE = 2
"""Old BSON binary subtype for binary data.

This is the old default subtype, the current
default is :data:`BINARY_SUBTYPE`.

.. versionadded:: 1.7
"""

UUID_SUBTYPE = 4
"""BSON binary subtype for a UUID.

This is the new BSON binary subtype for UUIDs. The
current default is :data:`OLD_UUID_SUBTYPE` but will
change to this in a future release.

.. versionchanged:: 2.1
   Changed to subtype 4.
.. versionadded:: 1.5
"""

OLD_UUID_SUBTYPE = 3
"""Old BSON binary subtype for a UUID.

:class:`uuid.UUID` instances will automatically be encoded
by :mod:`bson` using this subtype.

.. versionadded:: 2.1
"""

MD5_SUBTYPE = 5
"""BSON binary subtype for an MD5 hash.

.. versionadded:: 1.5
"""

USER_DEFINED_SUBTYPE = 128
"""BSON binary subtype for any user defined structure.

.. versionadded:: 1.5
"""


class Binary(str):
    """Representation of BSON binary data.

    This is necessary because we want to represent Python strings as
    the BSON string type. We need to wrap binary data so we can tell
    the difference between what should be considered binary data and
    what should be considered a string when we encode to BSON.

    Raises TypeError if `data` is not an instance of str or `subtype`
    is not an instance of int. Raises ValueError if `subtype` is not
    in [0, 256).

    :Parameters:
      - `data`: the binary data to represent
      - `subtype` (optional): the `binary subtype
        <http://bsonspec.org/#/specification>`_
        to use
    """

    def __new__(cls, data, subtype=BINARY_SUBTYPE):
        if not isinstance(data, str):
            raise TypeError("data must be an instance of str")
        if not isinstance(subtype, int):
            raise TypeError("subtype must be an instance of int")
        if subtype >= 256 or subtype < 0:
            raise ValueError("subtype must be contained in [0, 256)")
        self = str.__new__(cls, data)
        self.__subtype = subtype
        return self

    @property
    def subtype(self):
        """Subtype of this binary data.
        """
        return self.__subtype

    def __eq__(self, other):
        if isinstance(other, Binary):
            return (self.__subtype, str(self)) == (other.subtype, str(other))
        # We don't return NotImplemented here because if we did then
        # Binary("foo") == "foo" would return True, since Binary is a
        # subclass of str...
        return False

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "Binary(%s, %s)" % (str.__repr__(self), self.__subtype)


class UUIDLegacy(Binary):
    """UUID wrapper to support working with UUIDs stored as legacy
    BSON binary subtype 3.

    .. doctest::

      >>> import uuid
      >>> from bson.binary import Binary, UUIDLegacy, UUID_SUBTYPE
      >>> my_uuid = uuid.uuid4()
      >>> coll = db.test
      >>> coll.uuid_subtype = UUID_SUBTYPE
      >>> coll.insert({'uuid': Binary(my_uuid.bytes, 3)})
      ObjectId('...')
      >>> coll.find({'uuid': my_uuid}).count()
      0
      >>> coll.find({'uuid': UUIDLegacy(my_uuid)}).count()
      1
      >>> coll.find({'uuid': UUIDLegacy(my_uuid)})[0]['uuid']
      UUID('...')
      >>>
      >>> # Convert from subtype 3 to subtype 4
      >>> doc = coll.find_one({'uuid': UUIDLegacy(my_uuid)})
      >>> coll.save(doc)
      ObjectId('...')
      >>> coll.find({'uuid': UUIDLegacy(my_uuid)}).count()
      0
      >>> coll.find({'uuid': {'$in': [UUIDLegacy(my_uuid), my_uuid]}}).count()
      1
      >>> coll.find_one({'uuid': my_uuid})['uuid']
      UUID('...')

    Raises TypeError if `obj` is not an instance of :class:`~uuid.UUID`.

    :Parameters:
      - `obj`: An instance of :class:`~uuid.UUID`.
    """

    def __new__(cls, obj):
        if not isinstance(obj, UUID):
            raise TypeError("obj must be an instance of uuid.UUID")
        self = Binary.__new__(cls, obj.bytes, OLD_UUID_SUBTYPE)
        self.__uuid = obj
        return self

    @property
    def uuid(self):
        """UUID instance wrapped by this UUIDLegacy instance.
        """
        return self.__uuid

    def __repr__(self):
        return "UUIDLegacy('%s')" % self.__uuid



# Copyright 2009-2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exceptions raised by PyMongo."""

from bson.errors import *


class PyMongoError(Exception):
    """Base class for all PyMongo exceptions.

    .. versionadded:: 1.4
    """


class ConnectionFailure(PyMongoError):
    """Raised when a connection to the database cannot be made or is lost.
    """


class AutoReconnect(ConnectionFailure):
    """Raised when a connection to the database is lost and an attempt to
    auto-reconnect will be made.

    In order to auto-reconnect you must handle this exception, recognizing that
    the operation which caused it has not necessarily succeeded. Future
    operations will attempt to open a new connection to the database (and
    will continue to raise this exception until the first successful
    connection is made).
    """


class ConfigurationError(PyMongoError):
    """Raised when something is incorrectly configured.
    """


class OperationFailure(PyMongoError):
    """Raised when a database operation fails.

    .. versionadded:: 1.8
       The :attr:`code` attribute.
    """

    def __init__(self, error, code=None):
        self.code = code
        PyMongoError.__init__(self, error)


class TimeoutError(OperationFailure):
    """Raised when a database operation times out.

    .. versionadded:: 1.8
    """


class DuplicateKeyError(OperationFailure):
    """Raised when a safe insert or update fails due to a duplicate key error.

    .. note:: Requires server version **>= 1.3.0**

    .. versionadded:: 1.4
    """


class InvalidOperation(PyMongoError):
    """Raised when a client attempts to perform an invalid operation.
    """


class InvalidName(PyMongoError):
    """Raised when an invalid name is used.
    """


class CollectionInvalid(PyMongoError):
    """Raised when collection validation fails.
    """


class InvalidURI(ConfigurationError):
    """Raised when trying to parse an invalid mongodb URI.

    .. versionadded:: 1.5
    """

class UnsupportedOption(ConfigurationError):
    """Exception for unsupported options.

    .. versionadded:: 2.0
    """



#!/usr/bin/env python

class Handler(object):
    """
    Base class for data handlers
    """
    def handle(self, source):
        """
        Consume and handle a data source
        """
        raise NotImplementedError(
            "handle() has not been implemented")


# -*- coding: utf-8 -*-
#
# Copyright (C) 2007 Christopher Lenz
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

try:
    __version__ = __import__('pkg_resources').get_distribution('CouchDB').version
except:
    __version__ = '?'



#!/usr/bin/env python
"""

Copyright (c) 2007  Dustin Sallings <dustin@spy.net>
"""

import struct

# Command constants
CMD_GET = 0
CMD_SET = 1
CMD_SETQ = 0x11
CMD_ADD = 2
CMD_REPLACE = 3
CMD_DELETE = 4
CMD_INCR = 5
CMD_DECR = 6
CMD_QUIT = 7
CMD_FLUSH = 8
CMD_GETQ = 9
CMD_NOOP = 10
CMD_VERSION = 11
CMD_STAT = 0x10
CMD_APPEND = 0x0e
CMD_PREPEND = 0x0f
CMD_TOUCH = 0x1c
CMD_GAT = 0x1d
CMD_GET_REPLICA = 0x83

# SASL stuff
CMD_SASL_LIST_MECHS = 0x20
CMD_SASL_AUTH = 0x21
CMD_SASL_STEP = 0x22

# Bucket extension
CMD_CREATE_BUCKET = 0x85
CMD_DELETE_BUCKET = 0x86
CMD_LIST_BUCKETS = 0x87
CMD_EXPAND_BUCKET = 0x88
CMD_SELECT_BUCKET = 0x89

CMD_STOP_PERSISTENCE = 0x80
CMD_START_PERSISTENCE = 0x81
CMD_SET_FLUSH_PARAM = 0x82

CMD_SET_TAP_PARAM = 0x92
CMD_EVICT_KEY = 0x93

CMD_RESTORE_FILE = 0x98
CMD_RESTORE_ABORT = 0x99
CMD_RESTORE_COMPLETE = 0x9a

#Online update
CMD_START_ONLINEUPDATE = 0x9b
CMD_COMPLETE_ONLINEUPDATE = 0x9c
CMD_REVERT_ONLINEUPDATE = 0x9d

# TAP client registration
CMD_DEREGISTER_TAP_CLIENT = 0x9e

# Reset replication chain
CMD_RESET_REPLICATION_CHAIN = 0x9f

CMD_GET_META = 0xa0
CMD_GETQ_META = 0xa1

# Replication
CMD_TAP_CONNECT = 0x40
CMD_TAP_MUTATION = 0x41
CMD_TAP_DELETE = 0x42
CMD_TAP_FLUSH = 0x43
CMD_TAP_OPAQUE = 0x44
CMD_TAP_VBUCKET_SET = 0x45
CMD_TAP_CHECKPOINT_START = 0x46
CMD_TAP_CHECKPOINT_END = 0x47

# vbucket stuff
CMD_SET_VBUCKET_STATE = 0x3d
CMD_GET_VBUCKET_STATE = 0x3e
CMD_DELETE_VBUCKET = 0x3f

CMD_GET_LOCKED = 0x94

CMD_SYNC = 0x96

# event IDs for the SYNC command responses
CMD_SYNC_EVENT_PERSISTED  = 1
CMD_SYNC_EVENT_MODIFED    = 2
CMD_SYNC_EVENT_DELETED    = 3
CMD_SYNC_EVENT_REPLICATED = 4
CMD_SYNC_INVALID_KEY      = 5
CMD_SYNC_INVALID_CAS      = 6

VB_STATE_ACTIVE=1
VB_STATE_REPLICA=2
VB_STATE_PENDING=3
VB_STATE_DEAD=4
VB_STATE_NAMES={'active': VB_STATE_ACTIVE,
                'replica': VB_STATE_REPLICA,
                'pending': VB_STATE_PENDING,
                'dead': VB_STATE_DEAD}

# Parameter types of CMD_SET_PARAM command.
ENGINE_PARAM_FLUSH      = 1
ENGINE_PARAM_TAP        = 2
ENGINE_PARAM_CHECKPOINT = 3

COMMAND_NAMES = dict(((globals()[k], k) for k in globals() if k.startswith("CMD_")))

# TAP_OPAQUE types
TAP_OPAQUE_ENABLE_AUTO_NACK        = 0
TAP_OPAQUE_INITIAL_VBUCKET_STREAM  = 1
TAP_OPAQUE_ENABLE_CHECKPOINT_SYNC  = 2
TAP_OPAQUE_OPEN_CHECKPOINT         = 3

# TAP connect flags
TAP_FLAG_BACKFILL          = 0x01
TAP_FLAG_DUMP              = 0x02
TAP_FLAG_LIST_VBUCKETS     = 0x04
TAP_FLAG_TAKEOVER_VBUCKETS = 0x08
TAP_FLAG_SUPPORT_ACK       = 0x10
TAP_FLAG_REQUEST_KEYS_ONLY = 0x20
TAP_FLAG_CHECKPOINT        = 0x40
TAP_FLAG_REGISTERED_CLIENT = 0x80

TAP_FLAG_TYPES = {TAP_FLAG_BACKFILL: ">Q",
                  TAP_FLAG_REGISTERED_CLIENT: ">B"}

# TAP per-message flags
TAP_FLAG_ACK      = 0x01
TAP_FLAG_NO_VALUE = 0x02 # The value for the key is not included in the packet

# Flags, expiration
SET_PKT_FMT=">II"

# flags
GET_RES_FMT=">I"

# How long until the deletion takes effect.
DEL_PKT_FMT=""

## TAP stuff
# eng-specific length, flags, ttl, [res, res, res]; item flags, exp
TAP_MUTATION_PKT_FMT = ">HHbxxxII"
TAP_GENERAL_PKT_FMT = ">HHbxxx"

# amount, initial value, expiration
INCRDECR_PKT_FMT=">QQI"
# Special incr expiration that means do not store
INCRDECR_SPECIAL=0xffffffff
INCRDECR_RES_FMT=">Q"

# Time bomb
FLUSH_PKT_FMT=">I"

# Touch commands
# expiration
TOUCH_PKT_FMT=">I"
GAT_PKT_FMT=">I"
GETL_PKT_FMT=">I"

# 2 bit integer.  :/
VB_SET_PKT_FMT=">I"

MAGIC_BYTE = 0x80
REQ_MAGIC_BYTE = 0x80
RES_MAGIC_BYTE = 0x81

# magic, opcode, keylen, extralen, datatype, vbucket, bodylen, opaque, cas
REQ_PKT_FMT=">BBHBBHIIQ"
# magic, opcode, keylen, extralen, datatype, status, bodylen, opaque, cas
RES_PKT_FMT=">BBHBBHIIQ"
# min recv packet size
MIN_RECV_PACKET = struct.calcsize(REQ_PKT_FMT)
# The header sizes don't deviate
assert struct.calcsize(REQ_PKT_FMT) == struct.calcsize(RES_PKT_FMT)

EXTRA_HDR_FMTS={
    CMD_SET: SET_PKT_FMT,
    CMD_ADD: SET_PKT_FMT,
    CMD_REPLACE: SET_PKT_FMT,
    CMD_INCR: INCRDECR_PKT_FMT,
    CMD_DECR: INCRDECR_PKT_FMT,
    CMD_DELETE: DEL_PKT_FMT,
    CMD_FLUSH: FLUSH_PKT_FMT,
    CMD_TAP_MUTATION: TAP_MUTATION_PKT_FMT,
    CMD_TAP_DELETE: TAP_GENERAL_PKT_FMT,
    CMD_TAP_FLUSH: TAP_GENERAL_PKT_FMT,
    CMD_TAP_OPAQUE: TAP_GENERAL_PKT_FMT,
    CMD_TAP_VBUCKET_SET: TAP_GENERAL_PKT_FMT,
    CMD_SET_VBUCKET_STATE: VB_SET_PKT_FMT,
}

EXTRA_HDR_SIZES=dict(
    [(k, struct.calcsize(v)) for (k,v) in EXTRA_HDR_FMTS.items()])

ERR_NOT_FOUND = 0x01
ERR_EXISTS = 0x02
ERR_2BIG = 0x03
ERR_EINVAL = 0x04
ERR_NOT_STORED = 0x05
ERR_BAD_DELTA = 0x06
ERR_NOT_MY_VBUCKET = 0x07
ERR_AUTH = 0x20
ERR_AUTH_CONTINUE = 0x21
ERR_UNKNOWN_CMD = 0x81
ERR_ENOMEM = 0x82
ERR_NOT_SUPPORTED = 0x83
ERR_EINTERNAL = 0x84
ERR_EBUSY = 0x85
ERR_ETMPFAIL = 0x86

META_REVID = 0x01




from memcached.helper.data_helper import MemcachedClientHelper

class StatsCommon():

    @staticmethod
    def get_stats(servers, bucket, stat_param, stat):
        """Gets stats for a specific key from a list of servers

        Parameters:
            servers - A list of servers to get a stat from. ([TestInputServer])
            bucket - The name of the bucket to get a stat from. (String)
            stat_param - The parameter for the stats call, eg. 'checkpoint'. (String)
            stat - The name of the stat to get. (String)

        Returns:
            Dictionary - The key is the test input server and the value is the
            result of the for the stat passed in as a parameter.
        """
        result = {}
        for server in servers:
            client = MemcachedClientHelper.direct_client(server, bucket)
            stats = client.stats(stat_param)
            result[server] = stats[stat]
        return result



from tasks.future import Future
from tasks.taskmanager import TaskManager
from tasks.task import *


"""An API for scheduling tasks that run against Couchbase Server

This module is contains the top-level API's for scheduling and executing tasks. The
API provides a way to run task do syncronously and asynchronously.
"""

class Cluster(object):
    """An API for interacting with Couchbase clusters"""

    def __init__(self):
        self.task_manager = TaskManager()
        self.task_manager.start()

    def async_create_default_bucket(self, server, size, replicas=1):
        """Asynchronously creates the default bucket

        Parameters:
            server - The server to create the bucket on. (TestInputServer)
            size - The size of the bucket to be created. (int)
            replicas - The number of replicas for this bucket. (int)

        Returns:
            BucketCreateTask - A task future that is a handle to the scheduled task."""
        _task = BucketCreateTask(server, 'default', replicas, size)
        self.task_manager.schedule(_task)
        return _task

    def async_create_sasl_bucket(self, server, name, password, size, replicas):
        """Asynchronously creates a sasl bucket

        Parameters:
            server - The server to create the bucket on. (TestInputServer)
            name - The name of the bucket to be created. (String)
            password - The password for this bucket. (String)
            replicas - The number of replicas for this bucket. (int)
            size - The size of the bucket to be created. (int)

        Returns:
            BucketCreateTask - A task future that is a handle to the scheduled task."""
        _task = BucketCreateTask(server, name, replicas, size, password=password)
        self.task_manager.schedule(_task)
        return _task

    def async_create_standard_bucket(self, server, name, port, size, replicas):
        """Asynchronously creates a standard bucket

        Parameters:
            server - The server to create the bucket on. (TestInputServer)
            name - The name of the bucket to be created. (String)
            port - The port to create this bucket on. (String)
            replicas - The number of replicas for this bucket. (int)
            size - The size of the bucket to be created. (int)

        Returns:
            BucketCreateTask - A task future that is a handle to the scheduled task."""
        _task = BucketCreateTask(server, name, replicas, size, port)
        self.task_manager.schedule(_task)
        return _task

    def async_bucket_delete(self, server, bucket='default'):
        """Asynchronously deletes a bucket

        Parameters:
            server - The server to delete the bucket on. (TestInputServer)
            bucket - The name of the bucket to be deleted. (String)

        Returns:
            BucketDeleteTask - A task future that is a handle to the scheduled task."""
        _task = BucketDeleteTask(server, bucket)
        self.task_manager.schedule(_task)
        return _task

    def async_init_node(self, server):
        """Asynchronously initializes a node

        The task scheduled will initialize a nodes username and password and will establish
        the nodes memory quota to be 2/3 of the available system memory.

        Parameters:
            server - The server to initialize. (TestInputServer)

        Returns:
            NodeInitTask - A task future that is a handle to the scheduled task."""
        _task = NodeInitializeTask(server)
        self.task_manager.schedule(_task)
        return _task

    def async_load_gen_docs(self, server, bucket, generator, kv_store, op_type, exp = 0):
        _task = LoadDocumentsTask(server, bucket, generator, kv_store, op_type, exp)
        self.task_manager.schedule(_task)
        return _task

    def async_workload(self, server, bucket, kv_store, num_ops, create, read, update,
                       delete, exp):
        _task = WorkloadTask(server, bucket, kv_store, num_ops, create, read, update,
                             delete, exp)
        self.task_manager.schedule(_task)
        return _task

    def async_verify_data(self, server, bucket, kv_store):
        _task = ValidateDataTask(server, bucket, kv_store)
        self.task_manager.schedule(_task)
        return _task

    def async_verify_revid(self, src_server, dest_server, bucket, kv_store, ops_perf):
        _task = VerifyRevIdTask(src_server, dest_server, bucket, kv_store, ops_perf)
        self.task_manager.schedule(_task)
        return _task

    def async_rebalance(self, servers, to_add, to_remove):
        """Asyncronously rebalances a cluster

        Parameters:
            servers - All servers participating in the rebalance ([TestInputServers])
            to_add - All servers being added to the cluster ([TestInputServers])
            to_remove - All servers being removed from the cluster ([TestInputServers])

        Returns:
            RebalanceTask - A task future that is a handle to the scheduled task"""
        _task = RebalanceTask(servers, to_add, to_remove)
        self.task_manager.schedule(_task)
        return _task

    def async_wait_for_stats(self, servers, bucket, param, stat, comparison, value):
        """Asynchronously wait for stats

        Waits for stats to match the criteria passed by the stats variable. See
        couchbase.stats_tool.StatsCommon.build_stat_check(...) for a description of
        the stats structure and how it can be built.

        Parameters:
            servers - The servers to get stats from. Specifying multiple servers will
                cause the result from each server to be added together before
                comparing. ([TestInputServer])
            bucket - The name of the bucket (String)
            param - The stats parameter to use. (String)
            stat - The stat that we want to get the value from. (String)
            comparison - How to compare the stat result to the value specified.
            value - The value to compare to.

        Returns:
            RebalanceTask - A task future that is a handle to the scheduled task"""
        _task = StatsWaitTask(servers, bucket, param, stat, comparison, value)
        self.task_manager.schedule(_task)
        return _task

    def create_default_bucket(self, server, size, replicas=1, timeout=None):
        """Synchronously creates the default bucket

        Parameters:
            server - The server to create the bucket on. (TestInputServer)
            size - The size of the bucket to be created. (int)
            replicas - The number of replicas for this bucket. (int)

        Returns:
            boolean - Whether or not the bucket was created."""
        _task = self.async_create_default_bucket(server, size, replicas)
        return _task.result(timeout)

    def create_sasl_bucket(self, server, name, password, size, replicas, timeout=None):
        """Synchronously creates a sasl bucket

        Parameters:
            server - The server to create the bucket on. (TestInputServer)
            name - The name of the bucket to be created. (String)
            password - The password for this bucket. (String)
            replicas - The number of replicas for this bucket. (int)
            size - The size of the bucket to be created. (int)

        Returns:
            boolean - Whether or not the bucket was created."""
        _task = self.async_create_sasl_bucket(server, name, password, replicas, size)
        self.task_manager.schedule(_task)
        return _task.result(timeout)

    def create_standard_bucket(self, server, name, port, size, replicas, timeout=None):
        """Synchronously creates a standard bucket

        Parameters:
            server - The server to create the bucket on. (TestInputServer)
            name - The name of the bucket to be created. (String)
            port - The port to create this bucket on. (String)
            replicas - The number of replicas for this bucket. (int)
            size - The size of the bucket to be created. (int)

        Returns:
            boolean - Whether or not the bucket was created."""
        _task = self.async_create_standard_bucket(server, name, port, size, replicas)
        return _task.result(timeout)

    def bucket_delete(self, server, bucket='default', timeout=None):
        """Synchronously deletes a bucket

        Parameters:
            server - The server to delete the bucket on. (TestInputServer)
            bucket - The name of the bucket to be deleted. (String)

        Returns:
            boolean - Whether or not the bucket was deleted."""
        _task = self.async_bucket_delete(server, bucket)
        return _task.result(timeout)

    def init_node(self, server):
        """Synchronously initializes a node

        The task scheduled will initialize a nodes username and password and will establish
        the nodes memory quota to be 2/3 of the available system memory.

        Parameters:
            server - The server to initialize. (TestInputServer)

        Returns:
            boolean - Whether or not the node was properly initialized."""
        _task = self.async_init_node(server)
        return _task.result()

    def rebalance(self, servers, to_add, to_remove, timeout=None):
        """Syncronously rebalances a cluster

        Parameters:
            servers - All servers participating in the rebalance ([TestInputServers])
            to_add - All servers being added to the cluster ([TestInputServers])
            to_remove - All servers being removed from the cluster ([TestInputServers])

        Returns:
            boolean - Whether or not the rebalance was successful"""
        _task = self.async_rebalance(servers, to_add, to_remove)
        return _task.result(timeout)

    def load_gen_docs(self, server, bucket, generator, kv_store, op_type, exp = 0, timeout=None):
        _task = self.async_load_gen_docs(server, bucket, generator, kv_store, op_type, exp)
        return _task.result(timeout)

    def workload(self, server, bucket, kv_store, num_ops, create, read, update, delete, exp, timeout=None):
        _task = self.async_workload(server, bucket, kv_store, num_ops, create, read, update,
                                    delete, exp)
        return _task.result(timeout)

    def verify_data(self, server, bucket, kv_store, timeout=None):
        _task = self.async_verify_data(server, bucket, kv_store)
        return _task.result(timeout)

    def wait_for_stats(self, servers, bucket, param, stat, comparison, value, timeout=None):
        """Synchronously wait for stats

        Waits for stats to match the criteria passed by the stats variable. See
        couchbase.stats_tool.StatsCommon.build_stat_check(...) for a description of
        the stats structure and how it can be built.

        Parameters:
            servers - The servers to get stats from. Specifying multiple servers will
                cause the result from each server to be added together before
                comparing. ([TestInputServer])
            bucket - The name of the bucket (String)
            param - The stats parameter to use. (String)
            stat - The stat that we want to get the value from. (String)
            comparison - How to compare the stat result to the value specified.
            value - The value to compare to.

        Returns:
            boolean - Whether or not the correct stats state was seen"""
        _task = self.async_wait_for_stats(servers, bucket, param, stat, comparison, value)
        return _task.result(timeout)

    def shutdown(self, force=False):
        self.task_manager.shutdown(force)

    def async_create_view(self, server, design_doc_name, view, bucket = "default"):
        """Asynchronously creates a views in a design doc

        Parameters:
            server - The server to handle create view task. (TestInputServer)
            design_doc_name - Design doc to be created or updated with view(s) being created (String)
            view - The view being created (document.View)
            bucket - The name of the bucket containing items for this view. (String)

        Returns:
            ViewCreateTask - A task future that is a handle to the scheduled task."""
        _task = ViewCreateTask(server, design_doc_name, view, bucket)
        self.task_manager.schedule(_task)
        return _task

    def create_view(self, server, design_doc_name, view, bucket = "default", timeout=None):
        """Synchronously creates a views in a design doc

        Parameters:
            server - The server to handle create view task. (TestInputServer)
            design_doc_name - Design doc to be created or updated with view(s) being created (String)
            view - The view being created (document.View)
            bucket - The name of the bucket containing items for this view. (String)

        Returns:
            string - revision number of design doc."""
        _task = self.async_create_view(server, design_doc_name, view, bucket)
        return _task.result(timeout)

    def async_delete_view(self, server, design_doc_name, view, bucket = "default"):
        """Asynchronously deletes a views in a design doc

        Parameters:
            server - The server to handle delete view task. (TestInputServer)
            design_doc_name - Design doc to be deleted or updated with view(s) being deleted (String)
            view - The view being deleted (document.View)
            bucket - The name of the bucket containing items for this view. (String)

        Returns:
            ViewDeleteTask - A task future that is a handle to the scheduled task."""
        _task = ViewDeleteTask(server, design_doc_name, view, bucket)
        self.task_manager.schedule(_task)
        return _task

    def delete_view(self, server, design_doc_name, view, bucket = "default", timeout=None):
        """Synchronously deletes a views in a design doc

        Parameters:
            server - The server to handle delete view task. (TestInputServer)
            design_doc_name - Design doc to be deleted or updated with view(s) being deleted (String)
            view - The view being deleted (document.View)
            bucket - The name of the bucket containing items for this view. (String)

        Returns:
            boolean - Whether or not delete view was successful."""
        _task = self.async_delete_view(server, design_doc_name, view, bucket)
        return _task.result(timeout)


    def async_query_view(self, server, design_doc_name, view_name, query,
                         expected_rows = None, bucket = "default", retry_time = 2):
        """Asynchronously query a views in a design doc

        Parameters:
            server - The server to handle query view task. (TestInputServer)
            design_doc_name - Design doc with view(s) being queried(String)
            view_name - The view being queried (String)
            expected_rows - The number of rows expected to be returned from the query (int)
            bucket - The name of the bucket containing items for this view. (String)
            retry_time - The time in seconds to wait before retrying failed queries (int)

        Returns:
            ViewQueryTask - A task future that is a handle to the scheduled task."""
        _task = ViewQueryTask(server, design_doc_name, view_name, query, expected_rows, bucket, retry_time)
        self.task_manager.schedule(_task)
        return _task

    def query_view(self, server, design_doc_name, view_name, query,
                   expected_rows = None, bucket = "default", retry_time = 2, timeout=None):
        """Synchronously query a views in a design doc

        Parameters:
            server - The server to handle query view task. (TestInputServer)
            design_doc_name - Design doc with view(s) being queried(String)
            view_name - The view being queried (String)
            expected_rows - The number of rows expected to be returned from the query (int)
            bucket - The name of the bucket containing items for this view. (String)
            retry_time - The time in seconds to wait before retrying failed queries (int)

        Returns:
            ViewQueryTask - A task future that is a handle to the scheduled task."""
        _task = self.async_query_view(server, design_doc_name, view_name, query, expected_rows, bucket, retry_time)
        return _task.result(timeout)


    def modify_fragmentation_config(self, server, config, bucket = "default", timeout=None):
        """Synchronously modify fragmentation configuration spec

        Parameters:
            server - The server to handle fragmentation config task. (TestInputServer)
            config - New compaction configuration (dict - see task)
            bucket - The name of the bucket fragementation config applies to. (String)

        Returns:
            boolean - True if config values accepted."""

        _task = ModifyFragmentationConfigTask(server, config, bucket)
        self.task_manager.schedule(_task)
        return _task.result(timeout)


    def async_monitor_view_fragmentation(self, server,
                                         design_doc_name,
                                         fragmentation_value,
                                         bucket = "default",
                                         timeout = None):
        """Asynchronously monitor view fragmentation.

           When <fragmentation_value> is reached on the
           index file for <design_doc_name> this method
           will return.

        Parameters:
            server - The server to handle fragmentation config task. (TestInputServer)
            design_doc_name - design doc with views represented in index file. (String)
            fragmentation_value - target amount of fragmentation within index file to detect. (String)
            bucket - The name of the bucket design_doc belongs to. (String)

        Returns:
            MonitorViewFragmentationTask - A task future that is a handle to the scheduled task."""

        _task = MonitorViewFragmentationTask(server, design_doc_name,
                                             fragmentation_value, bucket)
        self.task_manager.schedule(_task)
        return _task
    def async_generate_expected_view_results(self, doc_generators, view, query):
        """Asynchronously generate expected view query results

        Parameters:
            doc_generators - Generators used for loading docs (DocumentGenerator[])
            view - The view with map function (View)
            query - Query params to filter docs from the generator. (dict)

        Returns:
            GenerateExpectedViewResultsTask - A task future that is a handle to the scheduled task."""

        _task = GenerateExpectedViewResultsTask(doc_generators, view, query)
        self.task_manager.schedule(_task)
        return _task

    def generate_expected_view_query_results(self, doc_generators, view, query, timeout = None):
        """Synchronously generate expected view query results

        Parameters:
            doc_generators - Generators used for loading docs (DocumentGenerator[])
            view - The view with map function (View)
            query - Query params to filter docs from the generator. (dict)

        Returns:
            list - A list of rows expected to be returned for given query"""

        _task = self.async_generate_expected_view_results(doc_generators, view, query)
        return _task.result(timeout)


    def async_view_query_verification(self, server, design_doc_name, view_name, query, expected_rows, num_verified_docs = 20, bucket = "default", query_timeout = 20):
        """Asynchronously query a views in a design doc and does full verification of results

        Parameters:
            server - The server to handle query verification task. (TestInputServer)
            design_doc_name - Design doc with view(s) being queried(String)
            view_name - The view being queried (String)
            query - Query params being used with the query. (dict)
            expected_rows - The number of rows expected to be returned from the query (int)
            num_verified_docs - The number of docs to verify that require memcached gets (int)
            bucket - The name of the bucket containing items for this view. (String)
            query_timeout - The time to allow a query with stale=false to run. (int)
            retry_time - The time in seconds to wait before retrying failed queries (int)

        Returns:
            ViewQueryVerificationTask - A task future that is a handle to the scheduled task."""
        _task = ViewQueryVerificationTask(server, design_doc_name, view_name, query, expected_rows, num_verified_docs, bucket, query_timeout)
        self.task_manager.schedule(_task)
        return _task

    def view_query_verification(self, server, design_doc_name, view_name, query, expected_rows, num_verified_docs = 20, bucket = "default", query_timeout = 20, timeout = None):
        """Synchronously query a views in a design doc and does full verification of results

        Parameters:
            server - The server to handle query verification task. (TestInputServer)
            design_doc_name - Design doc with view(s) being queried(String)
            view_name - The view being queried (String)
            query - Query params being used with the query. (dict)
            expected_rows - The number of rows expected to be returned from the query (int)
            num_verified_docs - The number of docs to verify that require memcached gets (int)
            bucket - The name of the bucket containing items for this view. (String)
            query_timeout - The time to allow a query with stale=false to run. (int)
            retry_time - The time in seconds to wait before retrying failed queries (int)

        Returns:
            dict - An object with keys: passed = True or False
                                        errors = reasons why verification failed """
        _task = self.async_view_query_verification(server, design_doc_name, view_name, query, expected_rows, num_verified_docs, bucket, query_timeout)
        return _task.result(timeout)


    def monitor_view_fragmentation(self, server,
                                   design_doc_name,
                                   fragmentation_value,
                                   bucket = "default",
                                   timeout = None):
        """Synchronously monitor view fragmentation.

           When <fragmentation_value> is reached on the
           index file for <design_doc_name> this method
           will return.

        Parameters:
            server - The server to handle fragmentation config task. (TestInputServer)
            design_doc_name - design doc with views represented in index file. (String)
            fragmentation_value - target amount of fragmentation within index file to detect. (String)
            bucket - The name of the bucket design_doc belongs to. (String)

        Returns:
            boolean - True if <fragmentation_value> reached"""

        _task = self.async_monitor_view_fragmentation(server, design_doc_name,
                                                      fragmentation_value,
                                                      bucket)
        self.task_manager.schedule(_task)
        return _task.result(timeout)

    def async_compact_view(self, server, design_doc_name, bucket = "default"):
        """Asynchronously run view compaction.

        Compacts index file represented by views within the specified <design_doc_name>

        Parameters:
            server - The server to handle fragmentation config task. (TestInputServer)
            design_doc_name - design doc with views represented in index file. (String)
            bucket - The name of the bucket design_doc belongs to. (String)

        Returns:
            ViewCompactionTask - A task future that is a handle to the scheduled task."""


        _task = ViewCompactionTask(server, design_doc_name, bucket)
        self.task_manager.schedule(_task)
        return _task

    def compact_view(self, server, design_doc_name, bucket = "default", timeout=None):
        """Synchronously run view compaction.

        Compacts index file represented by views within the specified <design_doc_name>

        Parameters:
            server - The server to handle fragmentation config task. (TestInputServer)
            design_doc_name - design doc with views represented in index file. (String)
            bucket - The name of the bucket design_doc belongs to. (String)

        Returns:
            boolean - True file size reduced after compaction, False if successful but no work done """

        _task = self.async_compact_view(server, design_doc_name, bucket)
        return _task.result(timeout)

    def async_failover(self, servers, to_failover):
        """Asyncronously fails over nodes

        Parameters:
            servers - All servers participating in the failover ([TestInputServers])
            to_failover - All servers being failed over ([TestInputServers])

        Returns:
            FailoverTask - A task future that is a handle to the scheduled task"""
        _task = FailoverTask(servers, to_failover)
        self.task_manager.schedule(_task)
        return _task

    def failover(self, servers, to_failover, timeout=None):
        """Syncronously fails over nodes

        Parameters:
            servers - All servers participating in the failover ([TestInputServers])
            to_failover - All servers being failed over ([TestInputServers])

        Returns:
            boolean - Whether or not the failover was successful"""
        _task = self.async_failover(servers, to_failover)
        return _task.result(timeout)



#!/usr/bin/env python

import re
import sys
import socket
import threading
import mcsoda
import errno

sys.path.append("lib")
sys.path.append(".")

try:
   import logging
   logging.config.fileConfig("mcsoda.logging.conf")
   log = logging.getLogger()
except:
   class P:
      def error(self, m): print(m)
      def info(self, m):  print(m)
   log = P()

class Reader(threading.Thread):
    def __init__(self, src, reader_go, reader_done, cur):

        self.src = src
        self.reader_go = reader_go
        self.reader_done = reader_done
        self.cur = cur
        self.inflight = 0
        self.received = 0
        threading.Thread.__init__(self)

    def run(self):
        self.reader_go.wait()
        self.reader_go.clear()
        while True:
            try:
                data = self.src.recv(16384)
                if not data:
                    log.error("[cbsoda] Reader.recv-nodata / reader recv'ed nothing: %s" % data)
                    self.cur["cur-Reader.recv-nodata"] = self.cur.get("cur-Reader.recv-nodata", 0) + 1
                    self.reader_done.set()
                    self.reader_go = None
                    self.src = None
                    return

                self.received += len(data)

                found = len(re.findall("HTTP", data))

                self.inflight -= found
            except Exception as e:
                log.error("[cbsoda] EXCEPTION: Reader.run: " + str(e))
                self.cur["cur-ex-Reader.run"] = self.cur.get("cur-ex-Reader.run", 0) + 1
                self.inflight = 0

            if self.inflight == 0:
                self.reader_done.set()
                self.reader_go.wait()
                self.reader_go.clear()


class StoreCouchbase(mcsoda.StoreMembaseBinary):


    def connect_host_port(self, host, port, user, pswd, bucket="default"):
        if not hasattr(self, 'identity'):
            self.identity = (port, user, pswd, bucket)

        super(StoreCouchbase, self).connect_host_port(host, *self.identity)
        self.memcacheds = self.awareness.memcacheds

        self.capi_skt = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.capi_skt.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        # We really need timeout. Otherwise reader hangs up from time to time
        self.capi_skt.settimeout(180)
        try:
            self.capi_host_port = (host, 8092)
            self.capi_skt.connect(self.capi_host_port)
        except:
            self.capi_host_port = (host, 9500)
            self.capi_skt.connect(self.capi_host_port)
        self.init_reader(self.capi_skt)

    def disconnect(self):
        """Close socket connection"""
        self.capi_skt.shutdown(2)
        self.capi_skt.close()

    def restart_connection(self):
        """Determine new hostname, drop existing connection, establish new
        socket connection."""
        servers = [host_port.split(':')[0] for host_port in self.memcacheds]
        host = servers[self.cfg.get('node_prefix', 0) % len(servers)]

        self.disconnect()
        self.connect_host_port(host, *self.identity)

    def init_reader(self, skt):
        self.reader_go = threading.Event()
        self.reader_done = threading.Event()
        self.reader = Reader(skt, self.reader_go, self.reader_done, self.cur)
        self.reader.daemon = True
        self.reader.start()

    def inflight_start(self):
        inflight_grp = mcsoda.StoreMembaseBinary.inflight_start(self)
        inflight_grp['queries'] = [] # Array of queries.
        return inflight_grp

    def inflight_complete(self, inflight_grp):
        arr = mcsoda.StoreMembaseBinary.inflight_complete(self, inflight_grp)
        # Returns tuple of...
        #   (array of tuples (memcached_server, buffer) for memcached sends,
        #    buffer for capi send).
        return arr, ''.join(inflight_grp['queries']), len(inflight_grp['queries'])

    def inflight_send(self, t):
        # Before sending new batch, check number of active nodes. Restart HTTP
        # connection if this number changes.
        if len(self.memcacheds) != len(self.awareness.memcacheds):
            self.memcacheds = self.awareness.memcacheds
            self.restart_connection()

        for_mc, buf_capi, num_capi = t

        sent_mc = mcsoda.StoreMembaseBinary.inflight_send(self, for_mc)

        sent_capi = len(buf_capi)
        if sent_capi > 0:
            try:
                self.reader.inflight += num_capi
                self.capi_skt.send(buf_capi)
            except socket.error, e:
                log.error("[cbsoda] EXCEPTION: StoreCouchbase.inflight_send-socket.error / capi_skt.send: " + str(e))
                self.cur["cur-ex-StoreCouchbase.inflight_send-socket.error"] = self.cur.get("cur-ex-StoreCouchbase.inflight_send-socket.error", 0) + 1
                if isinstance(e.args, tuple):
                    if e[0] == errno.EPIPE:
                        # Remote-end closed the socket - TODO: why does this happen?
                        self.capi_skt.close()
                        self.capi_skt = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        self.capi_skt.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                        self.capi_skt.connect(self.capi_host_port)
                        self.init_reader(self.capi_skt) # TODO: we lose some stats from the old reader?

                        # Resend data
                        self.reader.inflight = num_capi
                        self.capi_skt.send(buf_capi)

        return sent_mc + sent_capi

    def inflight_recv(self, inflight, inflight_grp, expectBuffer=None):
        r = self.reader.received

        num_capi = len(inflight_grp['queries'])
        if num_capi > 0:
            self.reader_go.set()

        received_mc = mcsoda.StoreMembaseBinary.inflight_recv(self,
                                                              inflight, inflight_grp,
                                                              expectBuffer=expectBuffer)

        if num_capi > 0:
            self.reader_done.wait()
            self.reader_done.clear()

        received_capi = self.reader.received - r

        return received_mc + received_capi

    def cmd_append(self, cmd, key_num, key_str, data, expiration, grp):
        if cmd[0] == 'q':
            self.cmds += 1
            m = grp['queries']

            prefix = "GET "
            suffix = " HTTP/1.1\r\n\r\n"

            # A sample queries array might be...
            #
            #   ["/default/_design/my_ddoc/_views/by_email?startkey={email}",
            #    "/default/_design/my_ddoc/_views/by_country?startkey={country}&limit=10"]
            #
            queries = self.cfg.get("queries", "")
            queries = queries.split(';')
            if len(queries) == 0 or len(queries[0]) == 0:
                queries = [ "/default/{key}" ]

            query = queries[self.cur.get("cur-queries", 0) % len(queries)]

            message = query.format(key_str,
                                   key=key_str,
                                   name=mcsoda.key_to_name(key_num, key_str).replace(" ", "+"),
                                   email=mcsoda.key_to_email(key_num, key_str),
                                   city=mcsoda.key_to_city(key_num, key_str),
                                   country=mcsoda.key_to_country(key_num, key_str),
                                   realm=mcsoda.key_to_realm(key_num, key_str),
                                   coins=mcsoda.key_to_coins(key_num, key_str),
                                   category=mcsoda.key_to_category(key_num, key_str),
                                   cmds=self.cmds,
                                   int10=self.cmds % 10,
                                   int100=self.cmds % 100,
                                   int1000=self.cmds % 1000)

            m.append(prefix + message + suffix)

            return 0, 0, 0, 0, 1

        return mcsoda.StoreMembaseBinary.cmd_append(self, cmd, key_num, key_str,
                                                    data, expiration, grp)


if __name__ == "__main__":
    extra_examples=["          %s couchbase://127.0.0.1:8092 ratio-queries=0.2",
                    "          %s couchbase://127.0.0.1:8092 ratio-queries=0.2 \\",
                    "               queries=/default/_design/DDOC/_view/by_email?startkey={email}",
                    "",
                    "Available keys for queries templates:",
                    "    key, name, email, city, country, realm, coins, category,",
                    "    cmds (the number of commands sent so far),",
                    "    int10, int100, int1000 (various sized integers)"
                    ]

    if len(sys.argv) >= 2 and \
       (sys.argv[1].find("couchbase") == 0 or \
        sys.argv[1].find("cb") == 0):
        mcsoda.main(sys.argv,
                    protocol="couchbase",
                    stores=[StoreCouchbase()],
                    extra_examples=extra_examples)
    else:
        mcsoda.main(sys.argv,
                    extra_examples=extra_examples)



#!/usr/bin/env python

#
#     Copyright 2010 Membase, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

# PYTHONPATH needs to be set up to point to mc_bin_client

import os
import subprocess

DEF_USERNAME = "Administrator"
DEF_PASSWORD = "password"
DEF_KIND = "json"

DEF_MOXI_PORT = 11211
DEF_HTTP_PORT = 8091
DEF_RAMSIZE = 256
DEF_REPLICA = 1

CLI_EXE_LOC = "../membase-cli/membase"
SSH_EXE_LOC = "/opt/membase/bin/cli/membase"

class CLIInterface(object):
    def __init__(self, server, http_port=DEF_HTTP_PORT, username=DEF_USERNAME, password=DEF_PASSWORD, kind=DEF_KIND, debug=False, ssh=False, sshkey=None):
        self.server = server
        self.http_port = http_port
        self.username = username
        self.password = password
        self.kind = kind
        self.debug = debug
        self.ssh = ssh
        self.sshkey = sshkey
        if (debug):
            self.acting_server_args = "-c %s:%d -u %s -p %s -o %s -d" % (self.server, self.http_port, self.username, self.password, self.kind)
        else:
            self.acting_server_args = "-c %s:%d -u %s -p %s -o %s" % (self.server, self.http_port, self.username, self.password, self.kind)

    def server_list(self):
        cmd = " server-list " + self.acting_server_args
        return self.execute_command(cmd)

    def server_info(self):
        cmd = " server-info " + self.acting_server_args
        return self.execute_command(cmd)

    def server_add(self, server_to_add, rebalance=False):
        if (rebalance):
            cmd = " rebalance " + self.acting_server_args + " --server-add=%s:%d --server-add-username=%s --server-add-password=%s"\
                % (server_to_add, self.http_port, self.username, self.password)
        else:
            cmd = " server-add " + self.acting_server_args + " --server-add=%s:%d --server-add-username=%s --server-add-password=%s"\
                % (server_to_add, self.http_port, self.username, self.password)
        return self.execute_command(cmd)

    def server_readd(self, server_to_readd):
        cmd = " server-readd " + self.acting_server_args + " --server-add=%s:%d --server-add-username=%s --server-add-password=%s"\
                % (server_to_readd, self.http_port, self.username, self.password)
        return self.execute_command(cmd)

    def rebalance(self):
        cmd = " rebalance " + self.acting_server_args
        return self.execute_command(cmd)

    def rebalance_stop(self):
        cmd = " reblance-stop " + self.acting_server_args
        return self.execute_command(cmd)

    def rebalance_status(self):
        cmd = " rebalance-status " + self.acting_server_args
        return self.execute_command(cmd)

    def failover(self, server_to_failover):
        cmd = " failover " + self.acting_server_args + " --server-failover %s" % (server_to_failover)
        return self.execute_command(cmd)
        
    def cluster_init(self, c_username=DEF_USERNAME, c_password=DEF_PASSWORD, c_port=DEF_HTTP_PORT, c_ramsize=DEF_RAMSIZE):
        cmd = " cluster-init " + self.acting_server_args\
            + " --cluster-init-username=%s --cluster-init-password=%s --cluster-init-port=%d --cluster-init-ramsize=%d"\
            % (c_username, c_password, c_port, c_ramsize)
        return self.execute_command(cmd)
        
    def node_init(self, path):
        cmd = " node-init " + self.acting_server_args + " --node-init-data-path=%s" % (path)
        return self.execute_command(cmd)
        
    def bucket_list(self):
        cmd = " bucket-list " + self.acting_server_args
        return self.execute_command(cmd)

    def bucket_create(self, bucket_name, bucket_type, bucket_port, bucket_password="", bucket_ramsize=DEF_RAMSIZE, replica_count=DEF_REPLICA):
        cmd = " bucket-create " + self.acting_server_args\
            + " --bucket=%s --bucket-type=%s --bucket-port=%d --bucket-password=%s --bucket-ramsize=%d --bucket-replica=%d"\
            % (bucket_name, bucket_type, bucket_port, bucket_password, bucket_ramsize, replica_count)
        return self.execute_command(cmd)
        
    def bucket_edit(self, bucket_name, bucket_type, bucket_port, bucket_password, bucket_ramsize, replica_count):
        cmd = " bucket-edit " + self.acting_server_args\
            + " --bucket=%s --bucket-type=%s --bucket-port=%d --bucket-password=%s --bucket-ramsize=%d --bucket-replica=%d"\
            % (bucket_name, bucket_type, bucket_port, bucket_password, bucket_ramsize, replica_count)
        return self.execute_command(cmd)
        
    def bucket_delete(self, bucket_name):
        cmd = " bucket-delete " + self.acting_server_args + " --bucket=%s" % (bucket_name)
        return self.execute_command(cmd)
        
    def bucket_flush(self):
        return "I don't work yet :-("

    def execute_command(self, cmd):
        if (self.ssh):
            return self.execute_ssh(SSH_EXE_LOC + cmd)
        else:
            return self.execute_local(CLI_EXE_LOC + cmd)

    def execute_local(self, cmd):
        rtn = ""
        process = subprocess.Popen(cmd ,shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        stdoutdata,stderrdata=process.communicate()
        rtn += stdoutdata
        return rtn

    def execute_ssh(self, cmd):
        rtn=""
        if (self.sshkey == None):
            process = subprocess.Popen("ssh root@%s \"%s\"" % (self.server,cmd),shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        else:
            process = subprocess.Popen("ssh -i %s root@%s \"%s\"" % (self.sshkey, self.server, cmd),shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        stdoutdata,stderrdata=process.communicate()
        rtn += stdoutdata
        return rtn



import time
import Queue

from threading import Thread
from old_tasks.task import Task

class TaskManager(Thread):
    def __init__(self):
        Thread.__init__(self)
        self.readyq = Queue.Queue()
        self.sleepq = Queue.Queue()
        self.running = True

    def schedule(self, task, sleep_time=0):
        if not isinstance(task, Task):
            raise TypeError("Tried to schedule somthing that's not a task")
        if sleep_time <= 0:
            self.readyq.put(task)
        else:
            wakeup_time = time.time() + sleep_time
            self.sleepq.put({'task': task, 'time': wakeup_time})

    def run(self):
        while (self.running == True or self.readyq.empty() != True or self.sleepq.empty() != True):
            if self.readyq.empty():
                time.sleep(1)
            else:
                task = self.readyq.get()
                task.step(self)
            for i in range(self.sleepq.qsize()):
                s_task = self.sleepq.get()
                if time.time() >= s_task['time']:
                    self.readyq.put(s_task['task'])
                else:
                    self.sleepq.put(s_task)


    def cancel(self):

        # stop queue processing
        self._Thread__stop()
        self.join()

        # empty task queue and cancel all tasks
        self._emptyq_and_cancel(self.readyq)
        self._emptyq_and_cancel(self.sleepq)

        self.stop()

    def _emptyq_and_cancel(self, queue):
        while not queue.empty():
            _t = queue.get()
            _t['task'].cancel()

    def stop(self):
        self.running = False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2007-2009 Christopher Lenz
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

"""Python client API for CouchDB.

>>> server = Server()
>>> db = server.create('python-tests')
>>> doc_id, doc_rev = db.save({'type': 'Person', 'name': 'John Doe'})
>>> doc = db[doc_id]
>>> doc['type']
'Person'
>>> doc['name']
'John Doe'
>>> del db[doc.id]
>>> doc.id in db
False

>>> del server['python-tests']
"""

import mimetypes
import os
from types import FunctionType
from inspect import getsource
from textwrap import dedent
import re
import warnings

from couchdb import http, json

__all__ = ['Server', 'Database', 'Document', 'ViewResults', 'Row']
__docformat__ = 'restructuredtext en'


DEFAULT_BASE_URL = os.environ.get('COUCHDB_URL', 'http://localhost:5984/')


class Server(object):
    """Representation of a CouchDB server.

    >>> server = Server()

    This class behaves like a dictionary of databases. For example, to get a
    list of database names on the server, you can simply iterate over the
    server object.

    New databases can be created using the `create` method:

    >>> db = server.create('python-tests')
    >>> db
    <Database 'python-tests'>

    You can access existing databases using item access, specifying the database
    name as the key:

    >>> db = server['python-tests']
    >>> db.name
    'python-tests'

    Databases can be deleted using a ``del`` statement:

    >>> del server['python-tests']
    """

    def __init__(self, url=DEFAULT_BASE_URL, full_commit=True, session=None):
        """Initialize the server object.

        :param url: the URI of the server (for example
                    ``http://localhost:5984/``)
        :param full_commit: turn on the X-Couch-Full-Commit header
        :param session: an http.Session instance or None for a default session
        """
        if isinstance(url, basestring):
            self.resource = http.Resource(url, session or http.Session())
        else:
            self.resource = url  # treat as a Resource object
        if not full_commit:
            self.resource.headers['X-Couch-Full-Commit'] = 'false'

    def __contains__(self, name):
        """Return whether the server contains a database with the specified
        name.

        :param name: the database name
        :return: `True` if a database with the name exists, `False` otherwise
        """
        try:
            self.resource.head(validate_dbname(name))
            return True
        except http.ResourceNotFound:
            return False

    def __iter__(self):
        """Iterate over the names of all databases."""
        status, headers, data = self.resource.get_json('_all_dbs')
        return iter(data)

    def __len__(self):
        """Return the number of databases."""
        status, headers, data = self.resource.get_json('_all_dbs')
        return len(data)

    def __nonzero__(self):
        """Return whether the server is available."""
        try:
            self.resource.head()
            return True
        except:
            return False

    def __repr__(self):
        return '<%s %r>' % (type(self).__name__, self.resource.url)

    def __delitem__(self, name):
        """Remove the database with the specified name.

        :param name: the name of the database
        :raise ResourceNotFound: if no database with that name exists
        """
        self.resource.delete_json(validate_dbname(name))

    def __getitem__(self, name):
        """Return a `Database` object representing the database with the
        specified name.

        :param name: the name of the database
        :return: a `Database` object representing the database
        :rtype: `Database`
        :raise ResourceNotFound: if no database with that name exists
        """
        db = Database(self.resource(name), validate_dbname(name))
        db.resource.head()  # actually make a request to the database
        return db

    def config(self):
        """The configuration of the CouchDB server.

        The configuration is represented as a nested dictionary of sections and
        options from the configuration files of the server, or the default
        values for options that are not explicitly configured.

        :rtype: `dict`
        """
        status, headers, data = self.resource.get_json('_config')
        return data

    def version(self):
        """The version string of the CouchDB server.

        Note that this results in a request being made, and can also be used
        to check for the availability of the server.

        :rtype: `unicode`"""
        status, headers, data = self.resource.get_json()
        return data['version']

    def stats(self, name=None):
        """Server statistics.

        :param name: name of single statistic, e.g. httpd/requests
                     (None -- return all statistics)
        """
        if not name:
            resource = self.resource('_stats')
        else:
            resource = self.resource('_stats', *name.split('/'))
        status, headers, data = resource.get_json()
        return data

    def tasks(self):
        """A list of tasks currently active on the server."""
        status, headers, data = self.resource.get_json('_active_tasks')
        return data

    def uuids(self, count=None):
        """Retrieve a batch of uuids

        :param count: a number of uuids to fetch
                      (None -- get as many as the server sends)
        :return: a list of uuids
        """
        if count is None:
            _, _, data = self.resource.get_json('_uuids')
        else:
            _, _, data = self.resource.get_json('_uuids', count=count)
        return data['uuids']

    def create(self, name):
        """Create a new database with the given name.

        :param name: the name of the database
        :return: a `Database` object representing the created database
        :rtype: `Database`
        :raise PreconditionFailed: if a database with that name already exists
        """
        self.resource.put_json(validate_dbname(name))
        return self[name]

    def delete(self, name):
        """Delete the database with the specified name.

        :param name: the name of the database
        :raise ResourceNotFound: if a database with that name does not exist
        :since: 0.6
        """
        del self[name]

    def replicate(self, source, target, **options):
        """Replicate changes from the source database to the target database.

        :param source: URL of the source database
        :param target: URL of the target database
        :param options: optional replication args, e.g. continuous=True
        """
        data = {'source': source, 'target': target}
        data.update(options)
        status, headers, data = self.resource.post_json('_replicator', data)
        return data


class Database(object):
    """Representation of a database on a CouchDB server.

    >>> server = Server()
    >>> db = server.create('python-tests')

    New documents can be added to the database using the `save()` method:

    >>> doc_id, doc_rev = db.save({'type': 'Person', 'name': 'John Doe'})

    This class provides a dictionary-like interface to databases: documents are
    retrieved by their ID using item access

    >>> doc = db[doc_id]
    >>> doc                 #doctest: +ELLIPSIS
    <Document '...'@... {...}>

    Documents are represented as instances of the `Row` class, which is
    basically just a normal dictionary with the additional attributes ``id`` and
    ``rev``:

    >>> doc.id, doc.rev     #doctest: +ELLIPSIS
    ('...', ...)
    >>> doc['type']
    'Person'
    >>> doc['name']
    'John Doe'

    To update an existing document, you use item access, too:

    >>> doc['name'] = 'Mary Jane'
    >>> db[doc.id] = doc

    The `save()` method creates a document with a random ID generated by
    CouchDB (which is not recommended). If you want to explicitly specify the
    ID, you'd use item access just as with updating:

    >>> db['JohnDoe'] = {'type': 'person', 'name': 'John Doe'}

    >>> 'JohnDoe' in db
    True
    >>> len(db)
    2

    >>> del server['python-tests']
    """

    def __init__(self, url, name=None, session=None):
        if isinstance(url, basestring):
            if not url.startswith('http'):
                url = DEFAULT_BASE_URL + url
            self.resource = http.Resource(url, session)
        else:
            self.resource = url
        self._name = name

    def __repr__(self):
        return '<%s %r>' % (type(self).__name__, self.name)

    def __contains__(self, id):
        """Return whether the database contains a document with the specified
        ID.

        :param id: the document ID
        :return: `True` if a document with the ID exists, `False` otherwise
        """
        try:
            _doc_resource(self.resource, id).head()
            return True
        except http.ResourceNotFound:
            return False

    def __iter__(self):
        """Return the IDs of all documents in the database."""
        return iter([item.id for item in self.view('_all_docs')])

    def __len__(self):
        """Return the number of documents in the database."""
        _, _, data = self.resource.get_json()
        return data['doc_count']

    def __nonzero__(self):
        """Return whether the database is available."""
        try:
            self.resource.head()
            return True
        except:
            return False

    def __delitem__(self, id):
        """Remove the document with the specified ID from the database.

        :param id: the document ID
        """
        resource = _doc_resource(self.resource, id)
        status, headers, data = resource.head()
        resource.delete_json(rev=headers['etag'].strip('"'))

    def __getitem__(self, id):
        """Return the document with the specified ID.

        :param id: the document ID
        :return: a `Row` object representing the requested document
        :rtype: `Document`
        """
        _, _, data = _doc_resource(self.resource, id).get_json()
        return Document(data)

    def __setitem__(self, id, content):
        """Create or update a document with the specified ID.

        :param id: the document ID
        :param content: the document content; either a plain dictionary for
                        new documents, or a `Row` object for existing
                        documents
        """
        resource = _doc_resource(self.resource, id)
        status, headers, data = resource.put_json(body=content)
        content.update({'_id': data['id'], '_rev': data['rev']})

    @property
    def name(self):
        """The name of the database.

        Note that this may require a request to the server unless the name has
        already been cached by the `info()` method.

        :rtype: basestring
        """
        if self._name is None:
            self.info()
        return self._name

    def create(self, data):
        """Create a new document in the database with a random ID that is
        generated by the server.

        Note that it is generally better to avoid the `create()` method and
        instead generate document IDs on the client side. This is due to the
        fact that the underlying HTTP ``POST`` method is not idempotent, and
        an automatic retry due to a problem somewhere on the networking stack
        may cause multiple documents being created in the database.

        To avoid such problems you can generate a UUID on the client side.
        Python (since version 2.5) comes with a ``uuid`` module that can be
        used for this::

            from uuid import uuid4
            doc_id = uuid4().hex
            db[doc_id] = {'type': 'person', 'name': 'John Doe'}

        :param data: the data to store in the document
        :return: the ID of the created document
        :rtype: `unicode`
        """
        warnings.warn('Database.create is deprecated, please use Database.save instead [2010-04-13]',
                      DeprecationWarning, stacklevel=2)
        _, _, data = self.resource.post_json(body=data)
        return data['id']

    def multisave(self, docs, **options):
        """Perform a bulk insertion of the given documents.
        """
        responses = []
        for doc in docs:
            id, rev = self.save(doc, **options)
            responses.append((id, rev))

    def save(self, doc, **options):
        """Create a new document or update an existing document.

        If doc has no _id then the server will allocate a random ID and a new
        document will be created. Otherwise the doc's _id will be used to
        identity the document to create or update. Trying to update an existing
        document with an incorrect _rev will raise a ResourceConflict exception.

        Note that it is generally better to avoid saving documents with no _id
        and instead generate document IDs on the client side. This is due to
        the fact that the underlying HTTP ``POST`` method is not idempotent,
        and an automatic retry due to a problem somewhere on the networking
        stack may cause multiple documents being created in the database.

        To avoid such problems you can generate a UUID on the client side.
        Python (since version 2.5) comes with a ``uuid`` module that can be
        used for this::

            from uuid import uuid4
            doc = {'_id': uuid4().hex, 'type': 'person', 'name': 'John Doe'}
            db.save(doc)

        :param doc: the document to store
        :param options: optional args, e.g. batch='ok'
        :return: (id, rev) tuple of the save document
        :rtype: `tuple`
        """
        if 'meta' in doc:
            func = _doc_resource(self.resource, doc['meta']['id']).put_json
            _, _, data = func(body=doc['json'], **options)
            return doc['meta']['id']
        else:
            func = self.resource.post_json
            _, _, data = func(body=doc, **options)
            return data['id']

    def cleanup(self):
        """Clean up old design document indexes.

        Remove all unused index files from the database storage area.

        :return: a boolean to indicate successful cleanup initiation
        :rtype: `bool`
        """
        headers = {'Content-Type': 'application/json'}
        _, _, data = self.resource('_view_cleanup').post_json(headers=headers)
        return data['ok']

    def commit(self):
        """If the server is configured to delay commits, or previous requests
        used the special ``X-Couch-Full-Commit: false`` header to disable
        immediate commits, this method can be used to ensure that any
        non-committed changes are committed to physical storage.
        """
        _, _, data = self.resource.post_json(
            '_ensure_full_commit',
            headers={'Content-Type': 'application/json'})
        return data

    def compact(self, ddoc=None):
        """Compact the database or a design document's index.

        Without an argument, this will try to prune all old revisions from the
        database. With an argument, it will compact the index cache for all
        views in the design document specified.

        :return: a boolean to indicate whether the compaction was initiated
                 successfully
        :rtype: `bool`
        """
        if ddoc:
            resource = self.resource('_compact', ddoc)
        else:
            resource = self.resource('_compact')
        _, _, data = resource.post_json(
            headers={'Content-Type': 'application/json'})
        return data['ok']

    def copy(self, src, dest):
        """Copy the given document to create a new document.

        :param src: the ID of the document to copy, or a dictionary or
                    `Document` object representing the source document.
        :param dest: either the destination document ID as string, or a
                     dictionary or `Document` instance of the document that
                     should be overwritten.
        :return: the new revision of the destination document
        :rtype: `str`
        :since: 0.6
        """
        if not isinstance(src, basestring):
            if not isinstance(src, dict):
                if hasattr(src, 'items'):
                    src = dict(src.items())
                else:
                    raise TypeError('expected dict or string, got %s' %
                                    type(src))
            src = src['_id']

        if not isinstance(dest, basestring):
            if not isinstance(dest, dict):
                if hasattr(dest, 'items'):
                    dest = dict(dest.items())
                else:
                    raise TypeError('expected dict or string, got %s' %
                                    type(dest))
            if '_rev' in dest:
                dest = '%s?%s' % (http.quote(dest['_id']),
                                  http.urlencode({'rev': dest['_rev']}))
            else:
                dest = http.quote(dest['_id'])

        _, _, data = self.resource._request('COPY', src,
                                            headers={'Destination': dest})
        data = json.decode(data.read())
        return data['rev']

    def delete(self, doc):
        """Delete the given document from the database.

        Use this method in preference over ``__del__`` to ensure you're
        deleting the revision that you had previously retrieved. In the case
        the document has been updated since it was retrieved, this method will
        raise a `ResourceConflict` exception.

        >>> server = Server()
        >>> db = server.create('python-tests')

        >>> doc = dict(type='Person', name='John Doe')
        >>> db['johndoe'] = doc
        >>> doc2 = db['johndoe']
        >>> doc2['age'] = 42
        >>> db['johndoe'] = doc2
        >>> db.delete(doc)
        Traceback (most recent call last):
          ...
        ResourceConflict: ('conflict', 'Document update conflict.')

        >>> del server['python-tests']

        :param doc: a dictionary or `Document` object holding the document data
        :raise ResourceConflict: if the document was updated in the database
        :since: 0.4.1
        """
        if doc['_id'] is None:
            raise ValueError('document ID cannot be None')
        _doc_resource(self.resource, doc['_id']).delete_json(rev=doc['_rev'])

    def get(self, id, default=None, **options):
        """Return the document with the specified ID.

        :param id: the document ID
        :param default: the default value to return when the document is not
                        found
        :return: a `Row` object representing the requested document, or `None`
                 if no document with the ID was found
        :rtype: `Document`
        """
        try:
            _, _, data = _doc_resource(self.resource, id).get_json(**options)
        except http.ResourceNotFound:
            return default
        if hasattr(data, 'items'):
            return Document(data)
        else:
            return data

    def revisions(self, id, **options):
        """Return all available revisions of the given document.

        :param id: the document ID
        :return: an iterator over Document objects, each a different revision,
                 in reverse chronological order, if any were found
        """
        try:
            resource = _doc_resource(self.resource, id)
            status, headers, data = resource.get_json(revs=True)
        except http.ResourceNotFound:
            return

        startrev = data['_revisions']['start']
        for index, rev in enumerate(data['_revisions']['ids']):
            options['rev'] = '%d-%s' % (startrev - index, rev)
            revision = self.get(id, **options)
            if revision is None:
                return
            yield revision

    def info(self, ddoc=None):
        """Return information about the database or design document as a
        dictionary.

        Without an argument, returns database information. With an argument,
        return information for the given design document.

        The returned dictionary exactly corresponds to the JSON response to
        a ``GET`` request on the database or design document's info URI.

        :return: a dictionary of database properties
        :rtype: ``dict``
        :since: 0.4
        """
        if ddoc is not None:
            _, _, data = self.resource('_design', ddoc, '_info').get_json()
        else:
            _, _, data = self.resource.get_json()
            self._name = data['db_name']
        return data

    def delete_attachment(self, doc, filename):
        """Delete the specified attachment.

        Note that the provided `doc` is required to have a ``_rev`` field.
        Thus, if the `doc` is based on a view row, the view row would need to
        include the ``_rev`` field.

        :param doc: the dictionary or `Document` object representing the
                    document that the attachment belongs to
        :param filename: the name of the attachment file
        :since: 0.4.1
        """
        resource = _doc_resource(self.resource, doc['_id'])
        _, _, data = resource.delete_json(filename, rev=doc['_rev'])
        doc['_rev'] = data['rev']

    def get_attachment(self, id_or_doc, filename, default=None):
        """Return an attachment from the specified doc id and filename.

        :param id_or_doc: either a document ID or a dictionary or `Document`
                          object representing the document that the attachment
                          belongs to
        :param filename: the name of the attachment file
        :param default: default value to return when the document or attachment
                        is not found
        :return: a file-like object with read and close methods, or the value
                 of the `default` argument if the attachment is not found
        :since: 0.4.1
        """
        if isinstance(id_or_doc, basestring):
            id = id_or_doc
        else:
            id = id_or_doc['_id']
        try:
            _, _, data = _doc_resource(self.resource, id).get(filename)
            return data
        except http.ResourceNotFound:
            return default

    def put_attachment(self, doc, content, filename=None, content_type=None):
        """Create or replace an attachment.

        Note that the provided `doc` is required to have a ``_rev`` field. Thus,
        if the `doc` is based on a view row, the view row would need to include
        the ``_rev`` field.

        :param doc: the dictionary or `Document` object representing the
                    document that the attachment should be added to
        :param content: the content to upload, either a file-like object or
                        a string
        :param filename: the name of the attachment file; if omitted, this
                         function tries to get the filename from the file-like
                         object passed as the `content` argument value
        :param content_type: content type of the attachment; if omitted, the
                             MIME type is guessed based on the file name
                             extension
        :since: 0.4.1
        """
        if filename is None:
            if hasattr(content, 'name'):
                filename = os.path.basename(content.name)
            else:
                raise ValueError('no filename specified for attachment')
        if content_type is None:
            content_type = ';'.join(
                filter(None, mimetypes.guess_type(filename))
            )

        resource = _doc_resource(self.resource, doc['_id'])
        status, headers, data = resource.put_json(filename, body=content, headers={
            'Content-Type': content_type
        }, rev=doc['_rev'])
        doc['_rev'] = data['rev']

    def query(self, map_fun, reduce_fun=None, language='javascript',
              wrapper=None, **options):
        """Execute an ad-hoc query (a "temp view") against the database.

        >>> server = Server()
        >>> db = server.create('python-tests')
        >>> db['johndoe'] = dict(type='Person', name='John Doe')
        >>> db['maryjane'] = dict(type='Person', name='Mary Jane')
        >>> db['gotham'] = dict(type='City', name='Gotham City')
        >>> map_fun = '''function(doc) {
        ...     if (doc.type == 'Person')
        ...         emit(doc.name, null);
        ... }'''
        >>> for row in db.query(map_fun):
        ...     print row.key
        John Doe
        Mary Jane

        >>> for row in db.query(map_fun, descending=True):
        ...     print row.key
        Mary Jane
        John Doe

        >>> for row in db.query(map_fun, key='John Doe'):
        ...     print row.key
        John Doe

        >>> del server['python-tests']

        :param map_fun: the code of the map function
        :param reduce_fun: the code of the reduce function (optional)
        :param language: the language of the functions, to determine which view
                         server to use
        :param wrapper: an optional callable that should be used to wrap the
                        result rows
        :param options: optional query string parameters
        :return: the view reults
        :rtype: `ViewResults`
        """
        return TemporaryView(self.resource('_temp_view'), map_fun,
                             reduce_fun, language=language,
                             wrapper=wrapper)(**options)

    def update(self, documents, **options):
        """Perform a bulk update or insertion of the given documents using a
        single HTTP request.

        >>> server = Server()
        >>> db = server.create('python-tests')
        >>> for doc in db.update([
        ...     Document(type='Person', name='John Doe'),
        ...     Document(type='Person', name='Mary Jane'),
        ...     Document(type='City', name='Gotham City')
        ... ]):
        ...     print repr(doc) #doctest: +ELLIPSIS
        (True, '...', '...')
        (True, '...', '...')
        (True, '...', '...')

        >>> del server['python-tests']

        The return value of this method is a list containing a tuple for every
        element in the `documents` sequence. Each tuple is of the form
        ``(success, docid, rev_or_exc)``, where ``success`` is a boolean
        indicating whether the update succeeded, ``docid`` is the ID of the
        document, and ``rev_or_exc`` is either the new document revision, or
        an exception instance (e.g. `ResourceConflict`) if the update failed.

        If an object in the documents list is not a dictionary, this method
        looks for an ``items()`` method that can be used to convert the object
        to a dictionary. Effectively this means you can also use this method
        with `mapping.Document` objects.

        :param documents: a sequence of dictionaries or `Document` objects, or
                          objects providing a ``items()`` method that can be
                          used to convert them to a dictionary
        :return: an iterable over the resulting documents
        :rtype: ``list``

        :since: version 0.2
        """
        docs = []
        for doc in documents:
            if isinstance(doc, dict):
                docs.append(doc)
            elif hasattr(doc, 'items'):
                docs.append(dict(doc.items()))
            else:
                raise TypeError('expected dict, got %s' % type(doc))

        content = options
        content.update(docs=docs)
        _, _, data = self.resource.post_json('_bulk_docs', body=content)

        results = []
        for idx, result in enumerate(data):
            if 'error' in result:
                if result['error'] == 'conflict':
                    exc_type = http.ResourceConflict
                else:
                    # XXX: Any other error types mappable to exceptions here?
                    exc_type = http.ServerError
                results.append((False, result['id'],
                                exc_type(result['reason'])))
            else:
                doc = documents[idx]
                if isinstance(doc, dict):  # XXX: Is this a good idea??
                    doc.update({'_id': result['id'], '_rev': result['rev']})
                results.append((True, result['id'], result['rev']))

        return results

    def purge(self, docs):
        """Perform purging (complete removing) of the given documents.

        Uses a single HTTP request to purge all given documents. Purged
        documents do not leave any meta-data in the storage and are not
        replicated.
        """
        content = {}
        for doc in docs:
            if isinstance(doc, dict):
                content[doc['_id']] = [doc['_rev']]
            elif hasattr(doc, 'items'):
                doc = dict(doc.items())
                content[doc['_id']] = [doc['_rev']]
            else:
                raise TypeError('expected dict, got %s' % type(doc))
        _, _, data = self.resource.post_json('_purge', body=content)
        return data

    def view(self, name, wrapper=None, **options):
        """Execute a predefined view.

        >>> server = Server()
        >>> db = server.create('python-tests')
        >>> db['gotham'] = dict(type='City', name='Gotham City')

        >>> for row in db.view('_all_docs'):
        ...     print row.id
        gotham

        >>> del server['python-tests']

        :param name: the name of the view; for custom views, use the format
                     ``design_docid/viewname``, that is, the document ID of the
                     design document and the name of the view, separated by a
                     slash
        :param wrapper: an optional callable that should be used to wrap the
                        result rows
        :param options: optional query string parameters
        :return: the view results
        :rtype: `ViewResults`
        """
        path = _path_from_name(name, '_view')
        return PermanentView(self.resource(*path), '/'.join(path),
                             wrapper=wrapper)(**options)

    def show(self, name, docid=None, **options):
        """Call a 'show' function.

        :param name: the name of the show function in the format
                     ``designdoc/showname``
        :param docid: optional ID of a document to pass to the show function.
        :param options: optional query string parameters
        :return: (headers, body) tuple, where headers is a dict of headers
                 returned from the show function and body is a readable
                 file-like instance
        """
        path = _path_from_name(name, '_show')
        if docid:
            path.append(docid)
        status, headers, body = self.resource(*path).get(**options)
        return headers, body

    def list(self, name, view, **options):
        """Format a view using a 'list' function.

        :param name: the name of the list function in the format
                     ``designdoc/listname``
        :param view: the name of the view in the format ``designdoc/viewname``
        :param options: optional query string parameters
        :return: (headers, body) tuple, where headers is a dict of headers
                 returned from the list function and body is a readable
                 file-like instance
        """
        path = _path_from_name(name, '_list')
        path.extend(view.split('/', 1))
        _, headers, body = _call_viewlike(self.resource(*path), options)
        return headers, body

    def update_doc(self, name, docid=None, **options):
        """Calls server side update handler.

        :param name: the name of the update handler function in the format
                     ``designdoc/updatename``.
        :param docid: optional ID of a document to pass to the update handler.
        :param options: optional query string parameters.
        :return: (headers, body) tuple, where headers is a dict of headers
                 returned from the list function and body is a readable
                 file-like instance
        """
        path = _path_from_name(name, '_update')
        if docid is None:
            func = self.resource(*path).post
        else:
            path.append(docid)
            func = self.resource(*path).put
        _, headers, body = func(**options)
        return headers, body

    def _changes(self, **opts):
        _, _, data = self.resource.get('_changes', **opts)
        lines = iter(data)
        for ln in lines:
            if not ln:  # skip heartbeats
                continue
            doc = json.decode(ln)
            if 'last_seq' in doc:  # consume the rest of the response if this
                for ln in lines:  # was the last line, allows conn reuse
                    pass
            yield doc

    def changes(self, **opts):
        """Retrieve a changes feed from the database.

        Takes since, feed, heartbeat and timeout options.
        """
        if opts.get('feed') == 'continuous':
            return self._changes(**opts)
        _, _, data = self.resource.get_json('_changes', **opts)
        return data


def _doc_resource(base, doc_id):
    """Return the resource for the given document id.
    """
    # Split an id that starts with a reserved segment, e.g. _design/foo, so
    # that the / that follows the 1st segment does not get escaped.
    if doc_id[:1] == '_':
        return base(*doc_id.split('/', 1))
    return base(doc_id)


def _path_from_name(name, type):
    """Expand a 'design/foo' style name to its full path as a list of
    segments.
    """
    if name.startswith('_'):
        return name.split('/')
    design, name = name.split('/', 1)
    return ['_design', design, type, name]


class Document(dict):
    """Representation of a document in the database.

    This is basically just a dictionary with the two additional properties
    `id` and `rev`, which contain the document ID and revision, respectively.
    """

    def __repr__(self):
        return '<%s %r@%r %r>' % (type(self).__name__, self.id, self.rev,
                                  dict([(k, v) for k, v in self.items()
                                        if k not in ('_id', '_rev')]))

    @property
    def id(self):
        """The document ID.

        :rtype: basestring
        """
        return self['_id']

    @property
    def rev(self):
        """The document revision.

        :rtype: basestring
        """
        return self['_rev']


class View(object):
    """Abstract representation of a view or query."""

    def __init__(self, url, wrapper=None, session=None):
        if isinstance(url, basestring):
            self.resource = http.Resource(url, session)
        else:
            self.resource = url
        self.wrapper = wrapper

    def __call__(self, **options):
        return ViewResults(self, options)

    def __iter__(self):
        return iter(self())

    def _exec(self, options):
        raise NotImplementedError


class PermanentView(View):
    """Representation of a permanent view on the server."""

    def __init__(self, uri, name, wrapper=None, session=None):
        View.__init__(self, uri, wrapper=wrapper, session=session)
        self.name = name

    def __repr__(self):
        return '<%s %r>' % (type(self).__name__, self.name)

    def _exec(self, options):
        _, _, data = _call_viewlike(self.resource, options)
        return data


class TemporaryView(View):
    """Representation of a temporary view."""

    def __init__(self, uri, map_fun, reduce_fun=None,
                 language='javascript', wrapper=None, session=None):
        View.__init__(self, uri, wrapper=wrapper, session=session)
        if isinstance(map_fun, FunctionType):
            map_fun = getsource(map_fun).rstrip('\n\r')
        self.map_fun = dedent(map_fun.lstrip('\n\r'))
        if isinstance(reduce_fun, FunctionType):
            reduce_fun = getsource(reduce_fun).rstrip('\n\r')
        if reduce_fun:
            reduce_fun = dedent(reduce_fun.lstrip('\n\r'))
        self.reduce_fun = reduce_fun
        self.language = language

    def __repr__(self):
        return '<%s %r %r>' % (type(self).__name__, self.map_fun,
                               self.reduce_fun)

    def _exec(self, options):
        body = {'map': self.map_fun, 'language': self.language}
        if self.reduce_fun:
            body['reduce'] = self.reduce_fun
        if 'keys' in options:
            options = options.copy()
            body['keys'] = options.pop('keys')
        content = json.encode(body).encode('utf-8')
        _, _, data = self.resource.post_json(body=content, headers={
            'Content-Type': 'application/json'
        }, **_encode_view_options(options))
        return data


def _encode_view_options(options):
    """Encode any items in the options dict that are sent as a JSON string to a
    view/list function.
    """
    retval = {}
    for name, value in options.items():
        if name in ('key', 'startkey', 'endkey') \
                or not isinstance(value, basestring):
            value = json.encode(value)
        retval[name] = value
    return retval


def _call_viewlike(resource, options):
    """Call a resource that takes view-like options.
    """
    if 'keys' in options:
        options = options.copy()
        keys = {'keys': options.pop('keys')}
        return resource.post_json(body=keys, **_encode_view_options(options))
    else:
        return resource.get_json(**_encode_view_options(options))


class ViewResults(object):
    """Representation of a parameterized view (either permanent or temporary)
    and the results it produces.

    This class allows the specification of ``key``, ``startkey``, and
    ``endkey`` options using Python slice notation.

    >>> server = Server()
    >>> db = server.create('python-tests')
    >>> db['johndoe'] = dict(type='Person', name='John Doe')
    >>> db['maryjane'] = dict(type='Person', name='Mary Jane')
    >>> db['gotham'] = dict(type='City', name='Gotham City')
    >>> map_fun = '''function(doc) {
    ...     emit([doc.type, doc.name], doc.name);
    ... }'''
    >>> results = db.query(map_fun)

    At this point, the view has not actually been accessed yet. It is accessed
    as soon as it is iterated over, its length is requested, or one of its
    `rows`, `total_rows`, or `offset` properties are accessed:

    >>> len(results)
    3

    You can use slices to apply ``startkey`` and/or ``endkey`` options to the
    view:

    >>> people = results[['Person']:['Person','ZZZZ']]
    >>> for person in people:
    ...     print person.value
    John Doe
    Mary Jane
    >>> people.total_rows, people.offset
    (3, 1)

    Use plain indexed notation (without a slice) to apply the ``key`` option.
    Note that as CouchDB makes no claim that keys are unique in a view, this
    can still return multiple rows:

    >>> list(results[['City', 'Gotham City']])
    [<Row id='gotham', key=['City', 'Gotham City'], value='Gotham City'>]

    >>> del server['python-tests']
    """

    def __init__(self, view, options):
        self.view = view
        self.options = options
        self._rows = self._total_rows = self._offset = None

    def __repr__(self):
        return '<%s %r %r>' % (type(self).__name__, self.view, self.options)

    def __getitem__(self, key):
        options = self.options.copy()
        if type(key) is slice:
            if key.start is not None:
                options['startkey'] = key.start
            if key.stop is not None:
                options['endkey'] = key.stop
            return ViewResults(self.view, options)
        else:
            options['key'] = key
            return ViewResults(self.view, options)

    def __iter__(self):
        return iter(self.rows)

    def __len__(self):
        return len(self.rows)

    def _fetch(self):
        data = self.view._exec(self.options)
        wrapper = self.view.wrapper or Row
        self._rows = [wrapper(row) for row in data['rows']]
        self._total_rows = data.get('total_rows')
        self._offset = data.get('offset', 0)

    @property
    def rows(self):
        """The list of rows returned by the view.

        :rtype: `list`
        """
        if self._rows is None:
            self._fetch()
        return self._rows

    @property
    def total_rows(self):
        """The total number of rows in this view.

        This value is `None` for reduce views.

        :rtype: `int` or ``NoneType`` for reduce views
        """
        if self._rows is None:
            self._fetch()
        return self._total_rows

    @property
    def offset(self):
        """The offset of the results from the first row in the view.

        This value is 0 for reduce views.

        :rtype: `int`
        """
        if self._rows is None:
            self._fetch()
        return self._offset


class Row(dict):
    """Representation of a row as returned by database views."""

    def __repr__(self):
        keys = 'id', 'key', 'error', 'value'
        items = ['%s=%r' % (k, self[k]) for k in keys if k in self]
        return '<%s %s>' % (type(self).__name__, ', '.join(items))

    @property
    def id(self):
        """The associated Document ID if it exists. Returns `None` when it
        doesn't (reduce results).
        """
        return self.get('id')

    @property
    def key(self):
        return self['key']

    @property
    def value(self):
        return self.get('value')

    @property
    def error(self):
        return self.get('error')

    @property
    def doc(self):
        """The associated document for the row. This is only present when the
        view was accessed with ``include_docs=True`` as a query parameter,
        otherwise this property will be `None`.
        """
        doc = self.get('doc')
        if doc:
            return Document(doc)


SPECIAL_DB_NAMES = set(['_users', "_replicator"])
VALID_DB_NAME = re.compile(r'^[a-z][a-z0-9_$()+-/]*$')


def validate_dbname(name):
    """ Validate dbname """
    if name in SPECIAL_DB_NAMES:
        return name
    if not VALID_DB_NAME.match(name):
        raise ValueError("Invalid database name '%s'" % name)
    return name



#!/usr/bin/env python

import struct

from obs_def import ObservePktFmt, OBS_OPCODE
from memcacheConstants import REQ_MAGIC_BYTE

class ObserveRequestKey:
    vbucket = 0x0000
    key_len = 0x0000
    key = ""

    def __init__(self, key, vbucket):
        self.key = key
        self.vbucket = vbucket
        self.key_len = len(key)

    def __repr__(self):
        return "<%s> vbucket: %d, key_len: %d, key: %s\n" %\
               (self.__class__.__name__, self.vbucket, self.key_len, self.key)

    def __str__(self):
        return self.__repr__()

    def __len__(self):
        return 4 + self.key_len

    def pack(self):
        pkt = []
        body = struct.pack(ObservePktFmt.OBS_REQ_BODY_FMT,
                           self.vbucket,
                           self.key_len)
        pkt.append(body)
        pkt.append(self.key)
        return ''.join(pkt)

class ObserveRequest:
    magic = REQ_MAGIC_BYTE
    opcode = OBS_OPCODE
    key_len = 0x0000
    extra_len = 0x00
    data_type = 0x00
    vbucket = 0x0000
    body_len = 0x00000000
    opaque = 0x00000000
    cas = 0x0000000000000000
    keys = []       # [:ObserveRequestKey]

    def __init__(self, keys, opaque=0x00000000, cas=0x0000000000000000):
        self.keys = keys
        self.opaque = opaque
        self.cas = cas

    def __repr__(self):
        repr = "<%s> magic: %x, opcode: %x, key_len: %d, extra_len: %x,"\
               " data_type: %x, vbucket: %d, body_len: %d, opaque: %s, cas: %x\n" %\
               (self.__class__.__name__, self.magic, self.opcode, self.key_len, self.extra_len,
                self.data_type, self.vbucket, self.body_len, self.opaque, self.cas)
        for key in self.keys:
            repr += str(key)
        return repr

    def __str__(self):
        return self.__repr__()

    def pack(self):
        pkt = []
        self.body_len = sum(len(s) for s in self.keys)

        header = struct.pack(ObservePktFmt.OBS_REQ_HDR_FMT, self.magic,
            self.opcode, self.key_len, self.extra_len,
            self.data_type, self.vbucket, self.body_len,
            self.opaque, self.cas)
        pkt.append(header)

        for key in self.keys:
            pkt.append(key.pack())

        return ''.join(pkt)



#  This software code is made available "AS IS" without warranties of any
#  kind.  You may copy, display, modify and redistribute the software
#  code either by itself or as incorporated into your code; provided that
#  you do not remove any proprietary notices.  Your use of this software
#  code is at your own risk and you waive any claim against Amazon Web
#  Services LLC or its affiliates with respect to your use of this software
#  code. (c) 2006 Amazon Web Services LLC or its affiliates.  All rights
#  reserved.

import base64
import hmac
import httplib
import re
import sys
import time
import urllib
import hashlib
# ElementTree is in stdlib from Python 2.5, so get it from there if we can:
try:
    from xml.etree import ElementTree as ET
except ImportError:
    from elementtree import ElementTree as ET

DEFAULT_HOST = 'ec2.amazonaws.com'
PORTS_BY_SECURITY = { True: 443, False: 80 }
API_VERSION = '2008-02-01'
RELEASE_VERSION = "22395"

class AWSAuthConnection(object):

    """
    Creates an authorized connection to EC2 containing wrappers for
    Query API calls.

    Each API call has a matching method on this class to perform the
    appropriate E2 action.

    @ivar verbose: Verbosity flag, defaults to false.  If set to true,
    some debug information is printed.

    """
    
    def __init__(self, aws_access_key_id, aws_secret_access_key,
                 is_secure=True, server=DEFAULT_HOST, port=None):

        if not port:
            port = PORTS_BY_SECURITY[is_secure]

        self.verbose = False
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        if (is_secure):
            self.connection = httplib.HTTPSConnection("%s:%d" % (server, port))
        else:
            self.connection = httplib.HTTPConnection("%s:%d" % (server, port))

    def pathlist(self, key, arr):
        """Converts a key and an array of values into AWS query param format."""
        params = {}
        i = 0
        for value in arr:
            i += 1
            params["%s.%s" % (key, i)] = value
        return params

    def allocate_address(self):
        """Makes an C{AllocateAddress} call.

        """
        params = {}
        return AllocateAddressResponse(self.make_request("AllocateAddress", params))

    def describe_addresses(self, publicIps=None):
        """Makes a C{DescribeAddresses} call.

        @param publicIps: List of addresses to describe.  If empty or
        omitted, all addresses will be described.

        """
        if publicIps == None: publicIps = []
        params = self.pathlist("PublicIp", publicIps)
        return DescribeAddressesResponse(self.make_request("DescribeAddresses", params))

    def release_address(self, publicIp):
        """Makes a C{ReleaseAddress} call.

        @param publicIp: Address to release

        """
        params = { "PublicIp": publicIp }
        return ReleaseAddressResponse(self.make_request("ReleaseAddress", params))

    def associate_address(self, publicIp, instanceId):
        """Makes an C{AssociateAddress} call.

        @param publicIp: Address to associate
        @param instanceId: Instance to associate address with

        """
        params = { "PublicIp": publicIp, "InstanceId": instanceId }
        return AssociateAddressResponse(self.make_request("AssociateAddress", params))

    def disassociate_address(self, publicIp):
        """Makes a C{DisassociateAddress} call.

        @param publicIp: Address to disassociate

        """
        params = { "PublicIp": publicIp }
        return DisassociateAddressResponse(self.make_request("DisassociateAddress", params))

    def describe_availability_zones(self, zoneNames=None):
        """Makes a C{DescribeAvailabilityZones} call.

        @param zoneNames: List of availability zones to describe.  If
        empty or omitted, all availability zones will be described.

        """
        if zoneNames == None: zoneNames = []
        params = self.pathlist("ZoneName", zoneNames)
        return DescribeAvailabilityZonesResponse(self.make_request("DescribeAvailabilityZones", params))


    def register_image(self, imageLocation):
        """Makes a C{RegisterImage} call.

        @param imageLocation: The location of the image manifest to
        register in S3.
        
        """
        params = { "ImageLocation": imageLocation }
        return RegisterImageResponse(self.make_request("RegisterImage", params))

    def describe_images(self, imageIds=None, owners=None, executableBy=None):
        """Makes a C{DescribeImages} call.

        @param imageIds: List of images to describe.  If empty or omitted, all
        images visible to the user are returned.

        @param owners: List of users to filter returned images by
        owner.  If empty or omitted, no filtering is done.

        @param executableBy: List of users (or user groups) to filter
        returned images by execution permissions.  If empty or
        omitted, no filtering is done.

        """
        if imageIds == None: imageIds = []
        if owners == None: owners = []
        if executableBy == None: executableBy = []
        params = self.pathlist("ImageId", imageIds)
        params.update(self.pathlist("Owner", owners))
        params.update(self.pathlist("ExecutableBy", executableBy))
        return DescribeImagesResponse(self.make_request("DescribeImages", params))

    def deregister_image(self, imageId):
        """Makes a C{DeregisterImage} call.

        @param imageId: The image id to deregister.

        """
        params = { "ImageId": imageId }
        return DeregisterImageResponse(self.make_request("DeregisterImage", params))

    def create_keypair(self, keyName):
        """Makes a C{CreateKeypair} call.

        @param keyName: Name for the new keypair.

        """
        params = { "KeyName": keyName }
        return CreateKeyPairResponse(self.make_request("CreateKeyPair", params))

    def describe_keypairs(self, keyNames=None):
        """Makes a C{DescribeKeypairs} call.

        @param keyNames: List of keypairs to describe.  If empty or
        omitted, all keypairs are returned.

        """
        if keyNames == None: keyNames = []
        params = self.pathlist("KeyName", keyNames)
        return DescribeKeyPairsResponse(self.make_request("DescribeKeyPairs", params))

    def delete_keypair(self, keyName):
        """Makes a C{DeleteKeypair} call.

        @param keyName: Name of keypair to delete.

        """

        params = { "KeyName": keyName }
        return DeleteKeyPairResponse(self.make_request("DeleteKeyPair", params))

    def run_instances(self, imageId, minCount=1, maxCount=1, keyName=None,
                      groupIds=None, userData=None, base64Encode=True,
                      addressingType=None, instanceType=None,
                      availabilityZone=None, kernelId=None, ramdiskId=None,
                      blockDeviceMapping=None):
        """Makes a C{RunInstances} call.

        @param imageId: AMI id to launch instances of.

        @param minCount: Minimum number of instances to launch.  If
        EC2 cannot launch at least this many, the call will fail.

        @param maxCount: Maximum number of instances to launch.  EC2
        will make a best-effort attempt to launch this many instances,
        but will not fail unless fewer than C{minCount} are available.

        @param keyName: Name of keypair to launch instances with.

        @param groupIds: List of security groups to launch instances
        in.

        @param userData: String containing user data to inject into
        launched instances.

        @param base64Encode: Specifies whether C{userData} string
        qshould be base64 encoded.  Defaults to True.

        @param addressingType: Specifies the address scheme to use for
        the instance. The supported (default) type is C{"public"}, or
        C{None} for the default. See the "Instance Addressing" section
        of the latest developer guide for more information.

        @param availabilityZone: Specifies the availability zone to
        launch in.

        @param kernelId: Kernel to launch the instances with.

        @param ramdiskId: Ramdisk to launch the instances with.

        @param blockDeviceMapping: Specifies the virtual block device
        mapping. This parameter is a list of two element lists or
        tuples. The first element is the virtual name and the second
        is the device name.

        """
        if groupIds == None: groupIds = []
        params = {
            "ImageId": imageId,
            "MinCount": str(minCount),
            "MaxCount": str(maxCount),
            }
        if addressingType is not None:
            params["AddressingType"] = addressingType
        params.update(self.pathlist("SecurityGroup", groupIds))
        if userData is not None:
            if base64Encode:
                userData = base64.encodestring(userData)
            params["UserData"] = userData
        if keyName is not None:
            params["KeyName"] = keyName
        if availabilityZone is not None:
            params["Placement.AvailabilityZone"] = availabilityZone
        if kernelId is not None:
            params["KernelId"] = kernelId
        if ramdiskId is not None:
            params["RamdiskId"] = ramdiskId
        if instanceType is not None:
            params["InstanceType"] = instanceType
        if blockDeviceMapping is not None:
            virtualNames, deviceNames = zip(*blockDeviceMapping)
            i = 0
            for value in arr:
                i += 1
                params["BlockDeviceMapping.%s.VirtualName" % (i)] = virtualNames[i]
                params["BlockDeviceMapping.%s.DeviceName" % (i)] = deviceNames[i]

        #return self.make_request("RunInstances", params)
        return RunInstancesResponseGetReservationId(self.make_request("RunInstances", params))
        #return RunInstancesResponse(self.make_request("RunInstances", params))

    def describe_instances(self, instanceIds=[]):
        """Makes a C{DescribeInstances} call.

        @param instanceIds: List of instances to describe.  If empty
        or omitted, all instances will be returned.

        """
        params = self.pathlist("InstanceId", instanceIds)
        return DescribeInstancesResponse(self.make_request("DescribeInstances", params))

    def get_console_output(self, instanceId):
        """Makes a C{GetConsoleOutput} call.

        @param instanceId: Instance from which to get console output.

        """
        params = { "InstanceId": instanceId }
        return GetConsoleOutputResponse(self.make_request("GetConsoleOutput", params))

    def reboot_instances(self, instanceIds):
        """Makes a C{RebootInstances} call.

        @param instanceIds: List of instances to reboot.

        """
        params = self.pathlist("InstanceId", instanceIds)
        return RebootInstancesResponse(self.make_request("RebootInstances", params))

    def terminate_instances(self, instanceIds):
        """Makes a C{TerminateInstances} call.

        @param instanceIds: List of instances to terminate.

        """
        params = self.pathlist("InstanceId", instanceIds)
        return TerminateInstancesResponse(self.make_request("TerminateInstances", params))

    def create_securitygroup(self, groupName, groupDescription):
        """Makes a C{CreateSecurityGroup} call.

        @param groupName: Name of group to create.

        @param groupDescription: Brief description of security group.

        """
        params = {
            "GroupName": groupName,
            "GroupDescription": groupDescription
            }
        return CreateSecurityGroupResponse(self.make_request("CreateSecurityGroup", params))

    def describe_securitygroups(self, groupNames=None):
        """Makes a C{DescribeSecurityGroups} call.

        @param groupNames: List of security groups to describe.  If
        empty or omitted, all security groups will be described.

        """
        if groupNames == None: groupNames = []
        params = self.pathlist("GroupName", groupNames)
        return DescribeSecurityGroupsResponse(self.make_request("DescribeSecurityGroups", params))

    def delete_securitygroup(self, groupName):
        """Makes a C{DeleteSecurityGroup} call.

        @param groupName: Name of security group to delete.

        """
        params = { "GroupName": groupName }
        return DeleteSecurityGroupResponse(self.make_request("DeleteSecurityGroup", params))

    def authorize(self, *args, **kwargs):
        """Makes an C{AuthorizeSecurityGroupIngress} call.

        L{authorize} and L{revoke} share parameter parsing code.
        See L{auth_revoke_impl} for parameters.  Also, see API docs
        for details of valid parameter combinations.

        """
        params = self.auth_revoke_impl(*args, **kwargs)
        return AuthorizeSecurityGroupIngressResponse(self.make_request("AuthorizeSecurityGroupIngress", params))

    def revoke(self, *args, **kwargs):
        """Makes an C{RevokeSecurityGroupIngress} call.

        L{authorize} and L{revoke} share parameter parsing code.
        See L{auth_revoke_impl} for parameters.  Also, see API docs
        for details of valid parameter combinations.

        """
        params = self.auth_revoke_impl(*args, **kwargs)
        return RevokeSecurityGroupIngressResponse(self.make_request("RevokeSecurityGroupIngress", params))

    def modify_image_attribute(self, imageId, attribute, operationType,
                               **kwargs):
        """Makes a C{ModifyImageAttribute} call.

        @param imageId: AMI to modify attribute of.

        @param attribute: Name of attribute to modify.

        @param operationType: Operation to perform on attribute.

        @param kwargs: Values for the attribute operation, documented below.

        @kwarg userIds: List of userIds (valid with
        C{'launchPermission'} attribute.)

        @kwarg userGroups: List of userGroups (valid with
        C{'launchPermission'} attribute.)

        """
        params = {
            "ImageId": imageId,
            "Attribute": attribute,
            "OperationType": operationType
            }
        if attribute == "launchPermission":
            if "userIds" in kwargs:
                params.update(self.pathlist("UserId", kwargs["userIds"]))
            if "userGroups" in kwargs:
                params.update(self.pathlist("UserGroup", kwargs["userGroups"]))
        elif attribute == "productCodes":
            if "productCodes" in kwargs:
                params.update(self.pathlist("ProductCode", kwargs["productCodes"]))
        return ModifyImageAttributeResponse(self.make_request("ModifyImageAttribute", params))
    
    def reset_image_attribute(self, imageId, attribute):
        """Makes a C{ResetImageAttribute} call.

        @param imageId: AMI to reset attribute of.

        @param attribute: Name of attribute to reset.

        """
        params = { "ImageId": imageId, "Attribute": attribute }
        return ResetImageAttributeResponse(self.make_request("ResetImageAttribute", params))
    
    def describe_image_attribute(self, imageId, attribute):
        """Makes a C{DescribeImageAttribute} call.

        @param imageId: AMI to describe attribute of.

        @param attribute: Name of attribute to describe.

        """
        params = { "ImageId": imageId, "Attribute": attribute }
        return DescribeImageAttributeResponse(self.make_request("DescribeImageAttribute", params))

    def auth_revoke_impl(self, groupName, ipProtocol=None, fromPort=None,
                         toPort=None, cidrIp=None,
                         sourceSecurityGroupName=None,
                         sourceSecurityGroupOwnerId=None):
        """Processes parameters for C{authorize} and C{revoke}.

        @param groupName: Name of security group to modify.

        @param ipProtocol: IP protocol in rule.  Valid vlaues are
        C{'tcp'}, C{'udp'} and C{'icmp'}.

        @param fromPort: Bottom of IP port range in rule.

        @param toPort: Top of IP port range in rule.

        @param cidrIp: CIDR IP range in rule.

        @param sourceSecurityGroupName: Security group name in rule.

        @param sourceSecurityGroupOwnerId: User id of security group
        in rule.

        """
        params = { "GroupName": groupName }
        if ipProtocol != None: params["IpProtocol"] = ipProtocol
        if fromPort != None: params["FromPort"] = str(fromPort)
        if toPort != None: params["ToPort"] = str(toPort)
        if cidrIp != None: params["CidrIp"] = cidrIp
        if sourceSecurityGroupName != None:
            params["SourceSecurityGroupName"] = sourceSecurityGroupName
        if sourceSecurityGroupOwnerId != None:
            params["SourceSecurityGroupOwnerId"] = sourceSecurityGroupOwnerId
        return params

    def confirm_product_instance(self, productCode, instanceId):
        """Makes a C{ConfirmProdcutInstance} call.

        @param productCode: The product code to confirm

        @param instanceId: The instance for which to confirm the product code.
        
        """
        params = { "ProcudtCode": productCode, "InstanceId": instanceId }
        return ConfirmProductInstanceResponse(self.make_request("ConfirmProductInstance", params))

    def make_request(self, action, params, data=''):
        params["Action"] = action
        if self.verbose:
            print params

        params["SignatureVersion"] = "1"
        params["AWSAccessKeyId"] = self.aws_access_key_id
        params["Version"] = API_VERSION
        params["Timestamp"] = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())

        params = zip(params.keys(), params.values())
        params.sort(key=lambda x: str.lower(x[0]))
        
        sig = self.get_aws_auth_param(params, self.aws_secret_access_key)

        path = "?%s&Signature=%s" % (
            "&".join(["=".join([param[0], urllib.quote_plus(param[1])]) for param in params]),
            sig)

        if self.verbose:
            print path

        headers = {
            'User-Agent': 'ec2-python-query 1.2-%s' % (RELEASE_VERSION)
            }

        self.connection.request("GET", "/%s" % path, data, headers)
        return self.connection.getresponse()

    def get_aws_auth_param(self, params, aws_secret_access_key):
        canonical_string = "".join(["".join(param) for param in params])
        return self.encode(aws_secret_access_key, canonical_string)

    # computes the base64'ed hmac-sha hash of the canonical string and
    # the secret access key, optionally urlencoding the result

    def encode(self, aws_secret_access_key, str, urlencode=True):
        b64_hmac = base64.encodestring(hmac.new(aws_secret_access_key, str, hashlib.sha1).digest()).strip()
        if urlencode:
            return urllib.quote_plus(b64_hmac)
        else:
            return b64_hmac



class Response(object):
    """Base class for XML response parsers.

    This class does everything except the API-call dependent parsing,
    which is handled in the child classes below.  Each child class
    should override the L{parse} method.
    
    """

    ERROR_XPATH = "Errors/Error"
    NAMESPACE = "http://ec2.amazonaws.com/doc/%s/" % (API_VERSION)

    def __init__(self, http_response):
        self.http_response = http_response
        self.http_xml = http_response.read()
        self.is_error = False
        if http_response.status == 200:
            self.structure = self.parse()
        else:
            self.is_error = True
            self.structure = self.parse_error()

    def parse_error(self):
        doc = ET.XML(self.http_xml)
        element = doc.find(self.ERROR_XPATH)
        errorCode = element.findtext("Code")
        errorMessage = element.findtext("Message")
        return [["%s: %s" % (errorCode, errorMessage)]]

    def parse(self):
        # Placeholder -- this method should be overridden in child classes.
        return None

    def __str__(self):
        return "\n".join(["\t".join(line) for line in self.structure])

    def fixxpath(self, xpath):
        # ElementTree wants namespaces in its xpaths, so here we add them.
        return "/".join(["{%s}%s" % (self.NAMESPACE, e) for e in xpath.split("/")])

    def findtext(self, element, xpath):
        return element.findtext(self.fixxpath(xpath))

    def findall(self, element, xpath):
        return element.findall(self.fixxpath(xpath))

    def find(self, element, xpath):
        return element.find(self.fixxpath(xpath))

    def fixnone(self, strnone):
        if strnone is None: return ""
        return strnone


class DescribeImagesResponse(Response):
    """Response parser class for C{DescribeImages} API call."""
    ELEMENT_XPATH = "imagesSet/item"
    def translate_isPublic(self, isPublic):
        return { "true": "public",
                 "false": "private" }[isPublic]
    
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []
        for element in self.findall(doc, self.ELEMENT_XPATH):
            lines.append(["IMAGE",
                          self.findtext(element, "imageId"),
                          self.findtext(element, "imageLocation"),
                          self.findtext(element, "imageOwnerId"),
                          self.findtext(element, "imageState"),
                          self.translate_isPublic(self.findtext(element, "isPublic")),
                          self.findtext(element, "architecture"),
                          self.findtext(element, "imageType"),
                          self.fixnone(self.findtext(element, "kernelId")),
                          self.fixnone(self.findtext(element, "ramdiskId")),
                          ])
        return lines


class RegisterImageResponse(Response):
    """Response parser class for C{RegisterImage} API call."""
    ELEMENT_XPATH = "imageId"
    def parse(self):
        doc = ET.XML(self.http_xml)
        return [["IMAGE", self.findtext(doc, self.ELEMENT_XPATH)]]


class DeregisterImageResponse(Response):
    """Response parser class for C{DeregisterImage} API call."""
    def parse(self):
        # If we don't get an error, the deregistration succeeded.
        return [["Image deregistered."]]


class CreateKeyPairResponse(Response):
    """Response parser class for C{CreateKeyPair} API call."""
    def parse(self):
        doc = ET.XML(self.http_xml)
        keyName = self.findtext(doc, "keyName")
        keyFingerprint = self.findtext(doc, "keyFingerprint")
        keyMaterial = self.findtext(doc, "keyMaterial")
        return [["KEYPAIR", keyName, keyFingerprint], [keyMaterial]]


class DescribeKeyPairsResponse(Response):
    """Response parser class for C{DescribeKeyPairs} API call."""
    ELEMENT_XPATH = "keySet/item"
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []
        for element in self.findall(doc, self.ELEMENT_XPATH):
            keyName = self.findtext(element, "keyName")
            keyFingerprint = self.findtext(element, "keyFingerprint")
            lines.append(["KEYPAIR", keyName, keyFingerprint])
        return lines


class DeleteKeyPairResponse(Response):
    """Response parser class for C{DeleteKeyPair} API call."""
    def parse(self):
        # If we don't get an error, the deletion succeeded.
        return [["Keypair deleted."]]


class InstanceSetResponse(Response):
    """ Response containing instance set items """
    def parseInstanceSet(self, root):
        """ Parse a set of instanceSet/item nodes """
        lines = []
        for element in self.findall(root, "instancesSet/item"):
            lines.append(["INSTANCE",
                          self.findtext(element, "instanceId"),
                          self.findtext(element, "imageId"),
                          self.findtext(element, "dnsName"),
                          self.findtext(element, "privateDnsName"),
                          self.findtext(element, "instanceState/name"),
                          self.fixnone(self.findtext(element, "keyName")),
                          self.findtext(element, "amiLaunchIndex"),
                          ",".join([p.text for p in self.findall(element, "productCodesSet/item/productCode")]),
                          self.findtext(element, "instanceType"),
                          self.findtext(element, "launchTime"),
                          self.findtext(element, "placement/availabilityZone"),
                          self.fixnone(self.findtext(element, "kernelId")),
                          self.fixnone(self.findtext(element, "ramdiskId")),
                          ])
        return lines

class RunInstancesResponse(InstanceSetResponse):
    """Response parser class for C{RunInstances} API call."""
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []
        reservationId = self.findtext(doc, "reservationId")
        ownerId = self.findtext(doc, "ownerId")
        groups = [g.text for g in self.findall(doc, "groupSet/item/groupId")]
        lines.append(["RESERVATION", reservationId, ownerId, ",".join(groups)])
        lines.extend(self.parseInstanceSet(doc))
        return lines

class DescribeInstancesResponse(InstanceSetResponse):
    """Response parser class for C{DescribeInstances} API call."""
    ELEMENT_XPATH = "reservationSet/item"
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []
        for rootelement in self.findall(doc, self.ELEMENT_XPATH):
            reservationId = self.findtext(rootelement, "reservationId")
            ownerId = self.findtext(rootelement, "ownerId")
            groups = [g.text for g in self.findall(rootelement, "groupSet/item/groupId")]
            lines.append(["RESERVATION", reservationId, ownerId, ",".join(groups)])
            lines.extend(self.parseInstanceSet(rootelement))
        return lines

    
class GetConsoleOutputResponse(Response):
    def parse(self):
        doc = ET.XML(self.http_xml)
        return [ [self.findtext(doc, "instanceId")],
                 [self.findtext(doc, "timestamp")],
                 [self.findtext(doc, "output")] ]


class TerminateInstancesResponse(Response):
    """Response parser class for C{TerminateInstances} API call."""
    ELEMENT_XPATH = "instancesSet/item"
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []
        for element in self.findall(doc, self.ELEMENT_XPATH):
            instanceId = self.findtext(element, "instanceId")
            shutdownState = self.findtext(element, "shutdownState/name")
            previousState = self.findtext(element, "previousState/name")
            # Only for debug mode, which we don't support yet:
            shutdownStateCode = self.findtext(element, "shutdownState/code")
            previousStateCode = self.findtext(element, "previousState/code")
            lines.append(["INSTANCE", instanceId, previousState, shutdownState])
        return lines


class RebootInstancesResponse(Response):
    """Response parser class for C{RebootInstances} API call."""
    def parse(self):
        # If we don't get an error, the creation succeeded.
        return [["Instances rebooted."]]


class CreateSecurityGroupResponse(Response):
    """Response parser class for C{CreateSecurityGroup} API call."""
    def parse(self):
        # If we don't get an error, the creation succeeded.
        return [["Security Group created."]]


class DescribeSecurityGroupsResponse(Response):
    """Response parser class for C{DescribeSecurityGroups} API call."""
    ELEMENT_XPATH = "securityGroupInfo/item"
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []
        for rootelement in self.findall(doc, self.ELEMENT_XPATH):
            groupName = self.findtext(rootelement, "groupName")
            ownerId = self.findtext(rootelement, "ownerId")
            groupDescription = self.findtext(rootelement, "groupDescription")
            lines.append(["GROUP", ownerId, groupName, groupDescription])
            for element in self.findall(rootelement, "ipPermissions/item"):
                ipProtocol = self.findtext(element, "ipProtocol")
                fromPort = self.findtext(element, "fromPort")
                toPort = self.findtext(element, "toPort")
                permList = [
                    "PERMISSION",
                    ownerId,
                    groupName,
                    "ALLOWS",
                    ipProtocol,
                    fromPort,
                    toPort,
                    "FROM"
                    ]
                for subelement in self.findall(element, "groups/item"):
                    userId = self.findtext(subelement, "userId")
                    targetGroupName = self.findtext(subelement, "groupName")
                    lines.append(permList + ["USER", userId, "GRPNAME", targetGroupName])
                for subelement in self.findall(element, "ipRanges/item"):
                    cidrIp = self.findtext(subelement, "cidrIp")
                    lines.append(permList + ["CIDR", cidrIp])
        return lines


class DeleteSecurityGroupResponse(Response):
    """Response parser class for C{DeleteSecurityGroup} API call."""
    def parse(self):
        # If we don't get an error, the deletion succeeded.
        return [["Security Group deleted."]]


class AuthorizeSecurityGroupIngressResponse(Response):
    """Response parser class for C{AuthorizeSecurityGroupIngress} API call."""
    def parse(self):
        # If we don't get an error, the authorization succeeded.
        return [["Ingress authorized."]]


class RevokeSecurityGroupIngressResponse(Response):
    """Response parser class for C{RevokeSecurityGroupIngress} API call."""
    def parse(self):
        # If we don't get an error, the revocation succeeded.
        return [["Ingress revoked."]]


class ModifyImageAttributeResponse(Response):
    """Response parser class for C{ModifyImageAttribute} API call."""
    def parse(self):
        # If we don't get an error, modification succeeded.
        return [["Image attribute modified."]]


class ResetImageAttributeResponse(Response):
    """Response parser class for C{ResetImageAttribute} API call."""
    def parse(self):
        # If we don't get an error, reset succeeded.
        return [["Image attribute reset."]]


class DescribeImageAttributeResponse(Response):
    """Response parser class for C{DescribeImageAttribute} API call."""
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []

        imageId = self.findtext(doc, "imageId")

        # Handle launchPermission attributes:
        element = self.find(doc, "launchPermission/item")
        if element != None:
            for subelement in element.getchildren():
                lines.append([
                    "launchPermission",
                    imageId,
                    subelement.tag.split("}")[1],
                    subelement.text
                    ])
        
        # Handle launchPermission attributes:
        element = self.find(doc, "productCodes/item")
        if element != None:
            for subelement in element.getchildren():
                lines.append([
                    "productCode",
                    imageId,
                    subelement.text
                    ])
        
        return lines

class AllocateAddressResponse(Response):
    """Response parser class for C{AllocateAddress} API call."""
    def parse(self):
        doc = ET.XML(self.http_xml)
        publicIp = self.findtext(doc, "publicIp")
        return [["ADDRESS", publicIp]]

class DescribeAddressesResponse(Response):
    """Response parser class for C{DescribeAddresses} API call."""
    ELEMENT_XPATH = "addressesSet/item"
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []
        for element in self.findall(doc, self.ELEMENT_XPATH):
            publicIp = self.findtext(element, "publicIp")
            instanceId = self.findtext(element, "instanceId")
            lines.append(["ADDRESS", publicIp, instanceId])
        return lines

class ReleaseAddressResponse(Response):
    """Response parser class for C{ReleaseAddress} API call."""
    def parse(self):
        # If we don't get an error, reset succeeded.
        return [["Address released."]]

class AssociateAddressResponse(Response):
    """Response parser class for C{AssociateAddress} API call."""
    def parse(self):
        # If we don't get an error, reset succeeded.
        return [["Address associated."]]

class DisassociateAddressResponse(Response):
    """Response parser class for C{DisassociateAddress} API call."""
    def parse(self):
        # If we don't get an error, reset succeeded.
        return [["Address disassociated."]]

class ConfirmProductInstanceResponse(Response):
    """Response parser class for C{ConfirmProductInstance} API call."""
    def parse(self):
        doc = ET.XML(self.http_xml)
        result = self.findtext(doc, "result")
        lines = [[result]]
        if result == "true":
            lines[0].append(self.findtext(doc, "ownerId"))

class DescribeAvailabilityZonesResponse(Response):
    """Response parser class for C{DescribeAvailabilityZones} API call."""
    ELEMENT_XPATH = "availabilityZoneInfo/item"
    def parse(self):
        doc = ET.XML(self.http_xml)
        lines = []
        for element in self.findall(doc, self.ELEMENT_XPATH):
            zoneName = self.findtext(element, "zoneName")
            zoneState = self.findtext(element, "zoneState")
            lines.append(["AVAILABILITYZONE", zoneName, zoneState])
        return lines



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2009 Christopher Lenz
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

"""Support for streamed reading and writing of multipart MIME content."""

from base64 import b64encode
from cgi import parse_header
try:
    from hashlib import md5
    md5; # Pyflakes workaround
except ImportError:
    from md5 import new as md5
import sys

__all__ = ['read_multipart', 'write_multipart']
__docformat__ = 'restructuredtext en'


CRLF = '\r\n'


def read_multipart(fileobj, boundary=None):
    """Simple streaming MIME multipart parser.
    
    This function takes a file-like object reading a MIME envelope, and yields
    a ``(headers, is_multipart, payload)`` tuple for every part found, where
    ``headers`` is a dictionary containing the MIME headers of that part (with
    names lower-cased), ``is_multipart`` is a boolean indicating whether the
    part is itself multipart, and ``payload`` is either a string (if
    ``is_multipart`` is false), or an iterator over the nested parts.
    
    Note that the iterator produced for nested multipart payloads MUST be fully
    consumed, even if you wish to skip over the content.
    
    :param fileobj: a file-like object
    :param boundary: the part boundary string, will generally be determined
                     automatically from the headers of the outermost multipart
                     envelope
    :return: an iterator over the parts
    :since: 0.5
    """
    headers = {}
    buf = []
    outer = in_headers = boundary is None

    next_boundary = boundary and '--' + boundary + '\n' or None
    last_boundary = boundary and '--' + boundary + '--\n' or None

    def _current_part():
        payload = ''.join(buf)
        if payload.endswith('\r\n'):
            payload = payload[:-2]
        elif payload.endswith('\n'):
            payload = payload[:-1]
        content_md5 = headers.get('content-md5')
        if content_md5:
            h = b64encode(md5(payload).digest())
            if content_md5 != h:
                raise ValueError('data integrity check failed')
        return headers, False, payload

    for line in fileobj:
        if in_headers:
            line = line.replace(CRLF, '\n')
            if line != '\n':
                name, value = line.split(':', 1)
                headers[name.lower().strip()] = value.strip()
            else:
                in_headers = False
                mimetype, params = parse_header(headers.get('content-type'))
                if mimetype.startswith('multipart/'):
                    sub_boundary = params['boundary']
                    sub_parts = read_multipart(fileobj, boundary=sub_boundary)
                    if boundary is not None:
                        yield headers, True, sub_parts
                        headers.clear()
                        del buf[:]
                    else:
                        for part in sub_parts:
                            yield part
                        return

        elif line.replace(CRLF, '\n') == next_boundary:
            # We've reached the start of a new part, as indicated by the
            # boundary
            if headers:
                if not outer:
                    yield _current_part()
                else:
                    outer = False
                headers.clear()
                del buf[:]
            in_headers = True

        elif line.replace(CRLF, '\n') == last_boundary:
            # We're done with this multipart envelope
            break

        else:
            buf.append(line)

    if not outer and headers:
        yield _current_part()


class MultipartWriter(object):

    def __init__(self, fileobj, headers=None, subtype='mixed', boundary=None):
        self.fileobj = fileobj
        if boundary is None:
            boundary = self._make_boundary()
        self.boundary = boundary
        if headers is None:
            headers = {}
        headers['Content-Type'] = 'multipart/%s; boundary="%s"' % (
            subtype, self.boundary
        )
        self._write_headers(headers)

    def open(self, headers=None, subtype='mixed', boundary=None):
        self.fileobj.write('--')
        self.fileobj.write(self.boundary)
        self.fileobj.write(CRLF)
        return MultipartWriter(self.fileobj, headers=headers, subtype=subtype,
                               boundary=boundary)

    def add(self, mimetype, content, headers=None):
        self.fileobj.write('--')
        self.fileobj.write(self.boundary)
        self.fileobj.write(CRLF)
        if headers is None:
            headers = {}
        if isinstance(content, unicode):
            ctype, params = parse_header(mimetype)
            if 'charset' in params:
                content = content.encode(params['charset'])
            else:
                content = content.encode('utf-8')
                mimetype = mimetype + ';charset=utf-8'
        headers['Content-Type'] = mimetype
        if content:
            headers['Content-Length'] = str(len(content))
            headers['Content-MD5'] = b64encode(md5(content).digest())
        self._write_headers(headers)
        if content:
            # XXX: throw an exception if a boundary appears in the content??
            self.fileobj.write(content)
            self.fileobj.write(CRLF)

    def close(self):
        self.fileobj.write('--')
        self.fileobj.write(self.boundary)
        self.fileobj.write('--')
        self.fileobj.write(CRLF)

    def _make_boundary(self):
        try:
            from uuid import uuid4
            return '==' + uuid4().hex + '=='
        except ImportError:
            from random import randrange
            token = randrange(sys.maxint)
            format = '%%0%dd' % len(repr(sys.maxint - 1))
            return '===============' + (format % token) + '=='

    def _write_headers(self, headers):
        if headers:
            for name in sorted(headers.keys()):
                self.fileobj.write(name)
                self.fileobj.write(': ')
                self.fileobj.write(headers[name])
                self.fileobj.write(CRLF)
        self.fileobj.write(CRLF)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def write_multipart(fileobj, subtype='mixed', boundary=None):
    r"""Simple streaming MIME multipart writer.

    This function returns a `MultipartWriter` object that has a few methods to
    control the nested MIME parts. For example, to write a flat multipart
    envelope you call the ``add(mimetype, content, [headers])`` method for
    every part, and finally call the ``close()`` method.

    >>> from StringIO import StringIO

    >>> buf = StringIO()
    >>> envelope = write_multipart(buf, boundary='==123456789==')
    >>> envelope.add('text/plain', 'Just testing')
    >>> envelope.close()
    >>> print buf.getvalue().replace('\r\n', '\n')
    Content-Type: multipart/mixed; boundary="==123456789=="
    <BLANKLINE>
    --==123456789==
    Content-Length: 12
    Content-MD5: nHmX4a6el41B06x2uCpglQ==
    Content-Type: text/plain
    <BLANKLINE>
    Just testing
    --==123456789==--
    <BLANKLINE>

    Note that an explicit boundary is only specified for testing purposes. If
    the `boundary` parameter is omitted, the multipart writer will generate a
    random string for the boundary.

    To write nested structures, call the ``open([headers])`` method on the
    respective envelope, and finish each envelope using the ``close()`` method:

    >>> buf = StringIO()
    >>> envelope = write_multipart(buf, boundary='==123456789==')
    >>> part = envelope.open(boundary='==abcdefghi==')
    >>> part.add('text/plain', 'Just testing')
    >>> part.close()
    >>> envelope.close()
    >>> print buf.getvalue().replace('\r\n', '\n') #:doctest +ELLIPSIS
    Content-Type: multipart/mixed; boundary="==123456789=="
    <BLANKLINE>
    --==123456789==
    Content-Type: multipart/mixed; boundary="==abcdefghi=="
    <BLANKLINE>
    --==abcdefghi==
    Content-Length: 12
    Content-MD5: nHmX4a6el41B06x2uCpglQ==
    Content-Type: text/plain
    <BLANKLINE>
    Just testing
    --==abcdefghi==--
    --==123456789==--
    <BLANKLINE>
    
    :param fileobj: a writable file-like object that the output should get
                    written to
    :param subtype: the subtype of the multipart MIME type (e.g. "mixed")
    :param boundary: the boundary to use to separate the different parts
    :since: 0.6
    """
    return MultipartWriter(fileobj, subtype=subtype, boundary=boundary)
